#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
The Coherence class stores the result of Series.wavelet_coherence(), whether WWZ or CWT.
It includes wavelet transform coherency and cross-wavelet transform.
"""
from ..utils import plotting
from ..utils import wavelet as waveutils
from ..utils import lipdutils

from ..core.scalograms import Scalogram, MultipleScalogram

import matplotlib.pyplot as plt
import numpy as np
from copy import deepcopy

from matplotlib.ticker import ScalarFormatter, FormatStrFormatter
from matplotlib import cm
from matplotlib import gridspec

from tqdm import tqdm
from scipy.stats.mstats import mquantiles
import warnings

def infer_period_unit_from_time_unit(time_unit):
    ''' infer a period unit based on the given time unit

    '''
    if time_unit is None:
        period_unit = None
    else:
        unit_group = lipdutils.timeUnitsCheck(time_unit)
        if unit_group != 'unknown':
            if unit_group == 'kage_units':
                period_unit = 'kyrs'
            else:
                period_unit = 'yrs'
        else:
            if time_unit[-1] == 's':
                period_unit = time_unit
            else:
                period_unit = f'{time_unit}s'

    return period_unit


class Coherence:
    '''Coherence object, meant to receive the WTC and XWT part of Series.wavelet_coherence()

    See also
    --------

    pyleoclim.core.series.Series.wavelet_coherence : Wavelet coherence method

    '''
    def __init__(self, frequency, scale, time, wtc, xwt, phase, coi=None,
                 wave_method=None, wave_args=None,
                 timeseries1=None, timeseries2=None, signif_qs=None, signif_method=None, qs =None,
                 freq_method=None, freq_kwargs=None, Neff_threshold=3, scale_unit=None, time_label=None):
        self.frequency = np.array(frequency)
        self.time = np.array(time)
        self.scale = np.array(scale)
        self.wtc = np.array(wtc)
        self.xwt = np.array(xwt)
        if coi is not None:
            self.coi = np.array(coi)
        else:
            self.coi = waveutils.make_coi(self.time, Neff_threshold=Neff_threshold)
        self.phase = np.array(phase)
        self.timeseries1 = timeseries1
        self.timeseries2 = timeseries2
        self.signif_qs = signif_qs
        self.signif_method = signif_method
        self.freq_method = freq_method
        self.freq_kwargs = freq_kwargs
        self.wave_method = wave_method
        if wave_args is not None:
            if 'freq' in wave_args.keys():
                wave_args['freq'] = np.array(wave_args['freq'])
            if 'tau' in wave_args.keys():
                wave_args['tau'] = np.array(wave_args['tau'])
        self.wave_args = wave_args
        self.qs        = qs

        if scale_unit is not None:
            self.scale_unit = scale_unit
        elif timeseries1 is not None:
            self.scale_unit = infer_period_unit_from_time_unit(timeseries1.time_unit)
        elif timeseries2 is not None:
            self.scale_unit = infer_period_unit_from_time_unit(timeseries2.time_unit)
        else:
            self.scale_unit = None

        if time_label is not None:
            self.time_label = time_label
        elif timeseries1 is not None:
            if timeseries1.time_unit is not None:
                self.time_label = f'{timeseries1.time_name} [{timeseries1.time_unit}]'
            else:
                self.time_label = f'{timeseries1.time_name}'
        elif timeseries2 is not None:
            if timeseries2.time_unit is not None:
                self.time_label = f'{timeseries2.time_name} [{timeseries2.time_unit}]'
            else:
                self.time_label = f'{timeseries2.time_name}'
        else:
            self.time_label = None

    def copy(self):
        '''Copy object
        '''
        return deepcopy(self)

    def plot(self, var='wtc', xlabel=None, ylabel=None, title='auto', figsize=[10, 8],
             ylim=None, xlim=None, in_scale=True, yticks=None, contourf_style={},
             phase_style={}, cbar_style={}, savefig_settings={}, ax=None,
             signif_clr='white', signif_linestyles='-', signif_linewidths=1,
             signif_thresh = 0.95, under_clr='ivory', over_clr='black', bad_clr='dimgray'):
        '''Plot the cross-wavelet results

        Parameters
        ----------
        
        var : str {'wtc', 'xwt'}
        
            variable to be plotted as color field. Default: 'wtc', the wavelet transform coherency.
            'xwt' plots the cross-wavelet transform instead.
        
        xlabel : str, optional
        
            x-axis label. The default is None.
        
        ylabel : str, optional
        
            y-axis label. The default is None.
        
        title : str, optional
        
            Title of the plot. The default is 'auto', where it is made from object metadata.
            To mute, pass title = None.
        
        figsize : list, optional
        
            Figure size. The default is [10, 8].
        
        ylim : list, optional
        
            y-axis limits. The default is None.
        
        xlim : list, optional
        
            x-axis limits. The default is None.
            
        in_scale : bool, optional
        
            
            Plots scales instead of frequencies The default is True.
        
        yticks : list, optional
        
            y-ticks label. The default is None.
        
        contourf_style : dict, optional
        
            Arguments for the contour plot. The default is {}.
        
        phase_style : dict, optional
        
            Arguments for the phase arrows. The default is {}. It includes:
            - 'pt': the default threshold above which phase arrows will be plotted
            - 'skip_x': the number of points to skip between phase arrows along the x-axis
            - 'skip_y':  the number of points to skip between phase arrows along the y-axis
            - 'scale': number of data units per arrow length unit (see matplotlib.pyplot.quiver)
            - 'width': shaft width in arrow units (see matplotlib.pyplot.quiver)
            - 'color': arrow color (see matplotlib.pyplot.quiver)
        
        cbar_style : dict, optional
        
            Arguments for the color bar. The default is {}.
       
        savefig_settings : dict, optional
        
            The default is {}.
            the dictionary of arguments for plt.savefig(); some notes below:
            - "path" must be specified; it can be any existed or non-existed path,
              with or without a suffix; if the suffix is not given in "path", it will follow "format"
            - "format" can be one of {"pdf", "eps", "png", "ps"}
        
        ax : ax, optional
        
            Matplotlib axis on which to return the figure. The default is None.
       
        signif_thresh: float in [0, 1]
        
            Significance threshold. Default is 0.95. If this quantile is not
            found in the qs field of the Coherence object, the closest quantile
            will be picked.
        
        signif_clr : str, optional
        
            Color of the significance line. The default is 'white'.
       
        signif_linestyles : str, optional
        
            Style of the significance line. The default is '-'.
        
        signif_linewidths : float, optional
        
            Width of the significance line. The default is 1.
       
        under_clr : str, optional
        
            Color for under 0. The default is 'ivory'.
        
        over_clr : str, optional
        
            Color for over 1. The default is 'black'.
       
        bad_clr : str, optional
        
            Color for missing values. The default is 'dimgray'.

        Returns
        -------
        
        fig, ax

        See also
        --------
        
        pyleoclim.core.coherence.Coherence.dashboard : plots a a dashboard showing the coherence and the cross-wavelet transform. 

        pyleoclim.core.series.Series.wavelet_coherence : computes the coherence from two timeseries.

        matplotlib.pyplot.quiver : quiver plot

        Examples
        --------

        Calculate the wavelet coherence of NINO3 and All India Rainfall and plot it:

        .. ipython:: python
            :okwarning:
            :okexcept:

            import pyleoclim as pyleo
            import pandas as pd
            import numpy as np
            data = pd.read_csv('https://raw.githubusercontent.com/LinkedEarth/Pyleoclim_util/Development/example_data/wtc_test_data_nino_even.csv')
            time = data['t'].values
            air = data['air'].values
            nino = data['nino'].values
            ts_air = pyleo.Series(time=time, value=air, time_name='Year (CE)',
                                  label='All India Rainfall', value_name='AIR (mm/month)')
            ts_nino = pyleo.Series(time=time, value=nino, time_name='Year (CE)',
                                   label='NINO3', value_name='NINO3 (K)')

            coh = ts_air.wavelet_coherence(ts_nino)
            @savefig coh_plot.png
            coh.plot()
            pyleo.closefig(fig)

        Establish significance against an AR(1) benchmark:

        .. ipython:: python
            :okwarning:
            :okexcept:

            coh_sig = coh.signif_test(number=20, qs=[.9,.95,.99])
            @savefig coh_sig_plot.png
            coh_sig.plot()
            pyleo.closefig(fig)

        Note that specifiying 3 significance thresholds does not take any more time as the quantiles are
        simply estimated from the same ensemble. By default, the plot function looks
        for the closest quantile to 0.95, but this is easy to adjust, e.g. for the 99th percentile:

        .. ipython:: python
            :okwarning:
            :okexcept:

            @savefig coh_sig_plot99.png
            coh_sig.plot(signif_thresh = 0.99)
            pyleo.closefig(fig)

        By default, the function plots the wavelet transform coherency (WTC), which quantifies where
        two timeseries exhibit similar behavior in time-frequency space, regardless of whether this
        corresponds to regions of high common power. To visualize the latter, you want to plot the
        cross-wavelet transform (XWT) instead, like so:

        .. ipython:: python
            :okwarning:
            :okexcept:

            @savefig xwt_plot.png
            coh_sig.plot(var='xwt')
            pyleo.closefig(fig)

        '''
        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)

        # handling NaNs
        mask_freq = []
        for i in range(np.size(self.frequency)):
            if all(np.isnan(self.wtc[:, i])):
                mask_freq.append(False)
            else:
                mask_freq.append(True)

        if in_scale:
            y_axis = self.scale[mask_freq]
            if ylabel is None:
                ylabel = f'Scale [{self.scale_unit}]' if self.scale_unit is not None else 'Scale'

            if yticks is None:
                yticks_default = np.array([0.1, 0.2, 0.5, 1, 2, 5, 10, 20, 50, 100, 200, 500, 1000, 2000, 5000, 1e4, 2e4, 5e4, 1e5, 2e5, 5e5, 1e6])
                mask = (yticks_default >= np.min(y_axis)) & (yticks_default <= np.max(y_axis))
                yticks = yticks_default[mask]
        else:
            y_axis = self.frequency[mask_freq]
            if ylabel is None:
                ylabel = f'Frequency [1/{self.scale_unit}]' if self.scale_unit is not None else 'Frequency'

        if signif_thresh > 1 or signif_thresh < 0:
            raise ValueError("The significance threshold must be in [0, 1] ")

        # plot color field for WTC or XWT
        contourf_args = {
            'cmap': 'magma',
            'origin': 'lower',
        }
        contourf_args.update(contourf_style)

        cmap = cm.get_cmap(contourf_args['cmap'])
        cmap.set_under(under_clr)
        cmap.set_over(over_clr)
        cmap.set_bad(bad_clr)
        contourf_args['cmap'] = cmap

        if var == 'wtc':
            lev = np.linspace(0, 1, 11)
            cont = ax.contourf(self.time, y_axis, self.wtc[:, mask_freq].T,
                               levels = lev, **contourf_args)
        elif var == 'xwt':
            cont = ax.contourf(self.time, y_axis, self.xwt[:, mask_freq].T,
                               levels = 11, **contourf_args) # just pass number of contours
        else:
            raise ValueError("Unknown variable; please choose either 'wtc' or 'xwt'")

        # plot significance levels
        if self.signif_qs is not None:
            signif_method_label = {
                'ar1': 'AR(1)',
            }
            if signif_thresh not in self.qs:
                isig = np.abs(np.array(self.qs) - signif_thresh).argmin()
                print("Significance threshold {:3.2f} not found in qs. Picking the closest, which is {:3.2f}".format(signif_thresh,self.qs[isig]))
            else:
                isig = self.qs.index(signif_thresh)

            if var == 'wtc':
                signif_coh = self.signif_qs[0].scalogram_list[isig] # extract WTC significance threshold
                signif_boundary = self.wtc[:, mask_freq].T / signif_coh.amplitude[:, mask_freq].T
            elif var == 'xwt':
                signif_coh = self.signif_qs[1].scalogram_list[isig] # extract XWT significance threshold
                signif_boundary = self.xwt[:, mask_freq].T / signif_coh.amplitude[:, mask_freq].T

            ax.contour(self.time, y_axis, signif_boundary, [-99, 1],
                       colors=signif_clr,
                       linestyles=signif_linestyles,
                       linewidths=signif_linewidths)
            if title is not None:
                ax.set_title("Lines:" + str(round(self.qs[isig]*100))+"% threshold")

        # plot colorbar
        cbar_args = {
            'label': var.upper(),
            'drawedges': False,
            'orientation': 'vertical',
            'fraction': 0.15,
            'pad': 0.05,
            'ticks': cont.levels
        }
        cbar_args.update(cbar_style)

        # assign colorbar to axis (instead of fig) : https://matplotlib.org/stable/gallery/subplots_axes_and_figures/colorbar_placement.html
        cb = plt.colorbar(cont, ax = ax, **cbar_args)

        # plot cone of influence
        ax.set_yscale('log')
        ax.plot(self.time, self.coi, 'k--')

        if ylim is None:
            ylim = [np.min(y_axis), np.min([np.max(y_axis), np.max(self.coi)])]

        ax.fill_between(self.time, self.coi, np.max(self.coi), color='white', alpha=0.5)

        if yticks is not None:
            ax.set_yticks(yticks)
            ax.yaxis.set_major_formatter(ScalarFormatter())
            ax.yaxis.set_major_formatter(FormatStrFormatter('%g'))

        if xlabel is None:
            xlabel = self.time_label

        if xlabel is not None:
            ax.set_xlabel(xlabel)

        if ylabel is not None:
            ax.set_ylabel(ylabel)

        # plot phase
        skip_x = np.max([int(np.size(self.time)//20), 1])
        skip_y = np.max([int(np.size(y_axis)//20), 1])

        phase_args = {'pt': 0.5, 'skip_x': skip_x, 'skip_y': skip_y,
                      'scale': 30, 'width': 0.004}
        phase_args.update(phase_style)

        pt = phase_args['pt']
        skip_x = phase_args['skip_x']
        skip_y = phase_args['skip_y']
        scale = phase_args['scale']
        width = phase_args['width']

        if 'color' in phase_style:
            color = phase_style['color']
        else:
            color = 'black'

        phase = np.copy(self.phase)[:, mask_freq]

        if self.signif_qs is None:
            if var == 'wtc':
                phase[self.wtc[:, mask_freq] < pt] = np.nan
            else:
                field = self.xwt[:, mask_freq]
                phase[field < pt*field.max()] = np.nan
        else:
            phase[signif_boundary.T < 1] = np.nan

        X, Y = np.meshgrid(self.time, y_axis)
        U, V = np.cos(phase).T, np.sin(phase).T

        ax.quiver(X[::skip_y, ::skip_x], Y[::skip_y, ::skip_x],
                  U[::skip_y, ::skip_x], V[::skip_y, ::skip_x],
                  scale=scale, width=width, zorder=99, color=color)

        ax.set_ylim(ylim)

        if xlim is not None:
            ax.set_xlim(xlim)

        lbl1 = self.timeseries1.label
        lbl2 = self.timeseries2.label

        if 'fig' in locals():
            if 'path' in savefig_settings:
                plotting.savefig(fig, settings=savefig_settings)
            if title is not None and  title != 'auto':
                fig.suptitle(title)
            elif title == 'auto' and lbl1 is not None and lbl1 is not None:
                title = 'Wavelet coherency ('+self.wave_method.upper() +') between '+ lbl1 + ' and ' + lbl2
                fig.suptitle(title)
            return fig, ax
        else:
            return ax


    def dashboard(self, title=None, figsize=[9,12], phase_style = {}, 
                  line_colors = ['tab:blue','tab:orange'], savefig_settings={},
                  ts_plot_kwargs = None, wavelet_plot_kwargs= None):
         ''' Cross-wavelet dashboard, including the two series, WTC and XWT.

             Note: this design balances many considerations, and is not easily customizable.

         Parameters
         ----------

         title : str, optional
         
             Title of the plot. The default is None.

         figsize : list, optional
         
             Figure size. The default is [9, 12], as this is an information-rich figure.
             
         line_colors : list, optional
         
             Colors for the 2 traces For nomenclature, see https://matplotlib.org/stable/gallery/color/named_colors.html

         savefig_settings : dict, optional
         
             The default is {}.
             the dictionary of arguments for plt.savefig(); some notes below:
             - "path" must be specified; it can be any existed or non-existed path,
               with or without a suffix; if the suffix is not given in "path", it will follow "format"
             - "format" can be one of {"pdf", "eps", "png", "ps"}

         phase_style : dict, optional
         
             Arguments for the phase arrows. The default is {}. It includes:
             - 'pt': the default threshold above which phase arrows will be plotted
             - 'skip_x': the number of points to skip between phase arrows along the x-axis
             - 'skip_y':  the number of points to skip between phase arrows along the y-axis
             - 'scale': number of data units per arrow length unit (see matplotlib.pyplot.quiver)
             - 'width': shaft width in arrow units (see matplotlib.pyplot.quiver)
             - 'color': arrow color (see matplotlib.pyplot.quiver)

         ts_plot_kwargs : dict
         
              arguments to be passed to the timeseries subplot, see pyleoclim.core.series.Series.plot for details

         wavelet_plot_kwargs : dict
         
              arguments to be passed to the contour subplots (XWT and WTC), [see pyleoclim.core.coherence.Coherence.plot for details]


         Returns
         -------
         
         fig, ax

         See also
         --------
         pyleoclim.core.coherence.Coherence.plot : creates a coherence plot

         pyleoclim.core.series.Series.wavelet_coherence : computes the coherence between two timeseries.

         pyleoclim.core.series.Series.plot: plots a timeseries

         matplotlib.pyplot.quiver: makes a quiver plot

         Examples
         --------

         Calculate the coherence of NINO3 and All India Rainfall and plot it as a dashboard:

         .. ipython:: python
             :okwarning:
             :okexcept:

             import pyleoclim as pyleo
             import pandas as pd
             data = pd.read_csv('https://raw.githubusercontent.com/LinkedEarth/Pyleoclim_util/Development/example_data/wtc_test_data_nino_even.csv')
             time = data['t'].values
             
             ts_air = pyleo.Series(time=time, value=data['air'].values, time_name='Year (CE)',
                                   label='All India Rainfall', value_name='AIR (mm/month)')
             ts_nino = pyleo.Series(time=time, value=data['nino'].values, time_name='Year (CE)',
                                    label='NINO3', value_name='NINO3 (K)')

             coh = ts_air.wavelet_coherence(ts_nino)
             coh_sig = coh.signif_test(number=10)

             @savefig coh_dash.png
             coh_sig.dashboard()
             pyleo.closefig(fig)
             
         You may customize colors like so:
             
         .. ipython:: python
             :okwarning:
             :okexcept:
             
             @savefig coh_dash1.png
             coh_sig.dashboard(line_colors=['teal','gold'])
             pyleo.closefig(fig)
             
         To export the figure, use `savefig_settings`: 
             
         .. ipython:: python
             :okwarning:
             :okexcept:
             
             coh_sig.dashboard(savefig_settings={'path':'coh_dash.png','dpi':300})
             pyleo.closefig(fig)

         '''
         # prepare options dictionaries
         savefig_settings = {} if savefig_settings is None else savefig_settings.copy()
         wavelet_plot_kwargs={} if wavelet_plot_kwargs is None else wavelet_plot_kwargs.copy()
         ts_plot_kwargs={} if ts_plot_kwargs is None else ts_plot_kwargs.copy()


         # create figure
         fig = plt.figure(figsize=figsize)
         gs = gridspec.GridSpec(8, 1)
         gs.update(wspace=0, hspace=0.5) # add some breathing room
         ax = {}

         # 1) plot timeseries
         #plt.rc('ytick', labelsize=8) 
         ax['ts1'] = plt.subplot(gs[0:2, 0])
         self.timeseries1.plot(ax=ax['ts1'], color=line_colors[0], **ts_plot_kwargs, legend=False)
         ax['ts1'].yaxis.label.set_color(line_colors[0])
         ax['ts1'].tick_params(axis='y', colors=line_colors[0],labelsize=8)
         ax['ts1'].spines['left'].set_color(line_colors[0])
         ax['ts1'].spines['bottom'].set_visible(False)
         ax['ts1'].grid(False)
         ax['ts1'].set_xlabel('')

         ax['ts2'] = ax['ts1'].twinx()
         self.timeseries2.plot(ax=ax['ts2'], color=line_colors[1],  **ts_plot_kwargs, legend=False)
         ax['ts2'].yaxis.label.set_color(line_colors[1])
         ax['ts2'].tick_params(axis='y', colors=line_colors[1],labelsize=8)
         ax['ts2'].spines['right'].set_color(line_colors[1])
         ax['ts2'].spines['right'].set_visible(True)
         ax['ts2'].spines['left'].set_visible(False)
         ax['ts2'].grid(False)

         # 2) plot WTC
         ax['wtc'] = plt.subplot(gs[2:5, 0], sharex=ax['ts1'])
         if 'cbar_style' not in wavelet_plot_kwargs:
             wavelet_plot_kwargs.update({'cbar_style':{'orientation': 'horizontal',
                                                       'pad': 0.15, 'aspect': 60}})
         self.plot(var='wtc',ax=ax['wtc'], title= None, **wavelet_plot_kwargs)
         #ax['wtc'].xaxis.set_visible(False)  # hide x axis
         ax['wtc'].set_xlabel('')

        # 3) plot XWT
         ax['xwt'] = plt.subplot(gs[5:8, 0], sharex=ax['ts1'])
         if 'phase_style' not in wavelet_plot_kwargs:
             wavelet_plot_kwargs.update({'phase_style':{'color': 'lightgray'}})
         self.plot(var='xwt',ax=ax['xwt'], title= None,
                   contourf_style={'cmap': 'viridis'},
                   cbar_style={'orientation': 'horizontal','pad': 0.2, 'aspect': 60},
                   phase_style=wavelet_plot_kwargs['phase_style'])

         #gs.tight_layout(fig) # this does nothing
         
         if 'fig' in locals():
             if 'path' in savefig_settings:
                 plotting.savefig(fig, settings=savefig_settings)
             return fig, ax
         else:
             return ax

    def signif_test(self, number=200, method='ar1sim', seed=None, qs=[0.95], settings=None, mute_pbar=False):
        '''Significance testing for Coherence objects

        The method obtains quantiles `qs` of the distribution of coherence between
        `number` pairs of Monte Carlo simulations of a process that resembles the original series.
        Currently, only AR(1) surrogates are supported.

        Parameters
        ----------
        
        number : int, optional
        
            Number of surrogate series to create for significance testing. The default is 200.
        
        method : {'ar1sim'}, optional
        
            Method through which to generate the surrogate series. The default is 'ar1sim'.
       
        seed : int, optional
        
            Fixes the seed for NumPy's random number generator.
            Useful for reproducibility. The default is None, so fresh, unpredictable
            entropy will be pulled from the operating system.
        
        qs : list, optional
        
            Significance levels to return. The default is [0.95].
            
        settings : dict, optional
        
            Parameters for surrogate model. The default is None.
            
        mute_pbar : bool, optional
        
            Mute the progress bar. The default is False.

        Returns
        -------
        
        new : pyleoclim.core.coherence.Coherence

            original Coherence object augmented with significance levels signif_qs,
            a list with the following `MultipleScalogram` objects:
            * 0: MultipleScalogram for the wavelet transform coherency (WTC)
            * 1: MultipleScalogram for the cross-wavelet transform (XWT)

            Each object contains as many Scalogram objects as qs contains values

        See also
        --------

        pyleoclim.core.series.Series.wavelet_coherence : Wavelet coherence

        pyleoclim.core.scalograms.Scalogram : Scalogram object

        pyleoclim.core.scalograms.MultipleScalogram : Multiple Scalogram object

        pyleoclim.core.coherence.Coherence.plot : plotting method for Coherence objects

        Examples
        --------

        Calculate the coherence of NINO3 and All India Rainfall and assess significance:

        .. ipython:: python
            :okwarning:
            :okexcept:

            import pyleoclim as pyleo
            import pandas as pd
            import numpy as np
            data = pd.read_csv('https://raw.githubusercontent.com/LinkedEarth/Pyleoclim_util/Development/example_data/wtc_test_data_nino_even.csv')
            time = data['t'].values
            air = data['air'].values
            nino = data['nino'].values
            ts_air = pyleo.Series(time=time, value=air, time_name='Year (CE)',
                                  label='All India Rainfall', value_name='AIR (mm/month)')
            ts_nino = pyleo.Series(time=time, value=nino, time_name='Year (CE)',
                                   label='NINO3', value_name='NINO3 (K)')

            coh = ts_air.wavelet_coherence(ts_nino)
            coh_sig = coh.signif_test(number=20)
            @savefig coh_sig_plot.png
            coh_sig.plot()
            pyleo.closefig(fig)

        By default, significance is assessed against a 95% benchmark derived from
        an AR(1) process fit to the data, using 200 Monte Carlo simulations.
        To customize, one can increase the number of simulations
        (more reliable, but slower), and the quantile levels.

        .. ipython:: python
            :okwarning:
            :okexcept:

            coh_sig2 = coh.signif_test(number=100, qs=[.9,.95,.99])
            @savefig coh_sig2_plot.png
            coh_sig2.plot()
            pyleo.closefig(fig)

        The plot() function will represent the 95% level as contours by default.
        If you need to show 99%, say, use the `signif_thresh` argument:

        .. ipython:: python
            :okwarning:
            :okexcept:

            @savefig coh_sig3_plot.png
            coh_sig2.plot(signif_thresh=0.99)
            pyleo.closefig(fig)

        Note that if the 99% quantile is not present, the plot method will look
        for the closest match, but lines are always labeled appropriately.
        For reproducibility purposes, it may be good to specify the (pseudo)random number
        generator's seed, like so:

        .. ipython:: python
            :okwarning:
            :okexcept:

            coh_sig27 = coh.signif_test(number=20, seed=27)

        This will generate exactly the same set of draws from the
        (pseudo)random number at every execution, which may be important for marginal features
        in small ensembles. In general, however, we recommend increasing the
        number of draws to check that features are robust.

        '''

        if number == 0:
            return self

        new = self.copy()
        surr1 = self.timeseries1.surrogates(
            number=number, seed=seed, method=method, settings=settings
        )
        surr2 = self.timeseries2.surrogates(
            number=number, seed=seed, method=method, settings=settings
        )

        wtcs, xwts = [], []

        for i in tqdm(range(number), desc='Performing wavelet coherence on surrogate pairs', total=number, disable=mute_pbar):
            coh_tmp = surr1.series_list[i].wavelet_coherence(surr2.series_list[i],
                                                             method  = self.wave_method,
                                                             settings = self.wave_args)
            wtcs.append(coh_tmp.wtc)
            xwts.append(coh_tmp.xwt)

        wtcs = np.array(wtcs)
        xwts = np.array(xwts)


        ne, nf, nt = np.shape(wtcs)

        # reshape because mquantiles only accepts inputs of at most 2D
        wtcs_r = np.reshape(wtcs, (ne, nf*nt))
        xwts_r = np.reshape(xwts, (ne, nf*nt))

        # define nd-arrays
        nq = len(qs)
        wtc_qs = np.ndarray(shape=(nq, nf, nt))
        xwt_qs = np.empty_like(wtc_qs)

        # for i in range(nf):
        #     for j in range(nt):
        #         wtc_qs[:,i,j] = mquantiles(wtcs[:,i,j], qs)
        #         xwt_qs[:,i,j] = mquantiles(xwts[:,i,j], qs)

        # extract quantiles and reshape
        wtc_qs = mquantiles(wtcs_r, qs, axis=0)
        wtc_qs = np.reshape(wtc_qs, (nq, nf, nt))
        xwt_qs = mquantiles(xwts_r, qs, axis=0)
        xwt_qs = np.reshape(xwt_qs, (nq, nf, nt))

        # put in Scalogram objects for export
        wtc_list, xwt_list = [],[]

        for i in range(nq):
            wtc_tmp = Scalogram(
                    frequency=self.frequency, time=self.time, amplitude=wtc_qs[i,:,:],
                    coi=self.coi, scale = self.scale,
                    freq_method=self.freq_method, freq_kwargs=self.freq_kwargs, label=f'{qs[i]*100:g}%',
                )
            wtc_list.append(wtc_tmp)
            xwt_tmp = Scalogram(
                    frequency=self.frequency, time=self.time, amplitude=xwt_qs[i,:,:],
                    coi=self.coi, scale = self.scale,
                    freq_method=self.freq_method, freq_kwargs=self.freq_kwargs, label=f'{qs[i]*100:g}%',
                )

            xwt_list.append(xwt_tmp)

        new.signif_qs = []
        new.signif_qs.append(MultipleScalogram(scalogram_list=wtc_list)) # Export WTC quantiles
        new.signif_qs.append(MultipleScalogram(scalogram_list=xwt_list)) # Export XWT quantiles
        new.signif_method = method
        new.qs = qs

        return new
    
    def phase_stats(self, scales, number=1000, level=0.05):
        ''' Estimate phase angle statistics of a Coherence object

        As per [1], the strength (consistency) of a phase relationship is assessed using:
            
        * sigma, the circular standard deviation   
        
        * kappa,  an estimate of the Von Mises distribution's concentration parameter. 
            It is a reciprocal measure of dispersion, so 1/kappa is analogous to the variance) [3].
        
        Because of inherent persistence of geophysical signals and of the 
        reproducing kernel of the continuous wavelet transform [3], phase statistics are 
        assessed relative to an AR(1) model fit to the angle deviations observed at the requested scale(s). 
    
        Specifically, if `number` is specified, the method simulates `number` 
        Monte Carlo realizations of an AR(1) process fit to fluctuations around 
        the mean angle. This ensemble is used to obtain the confidence limits: 
        `sigma_lo` (`level` quantile) and `kappa_hi` (1-`level` quantile). 
        These correspond to 1-tailed tests of the strength of the relationship.

        Parameters
        ----------
        
        scales : float
        
            scale at which to evaluate the phase angle
        
        number : int, optional
        
            number of AR(1) series to create for significance testing. The default is 1000.
       
        level : float, optional
        
           significance level against which to gauge sigma and kappa. default: 0.05
        

        Returns
        -------
        
        result : dict
        
            contains angle_mean (the mean angle for those scales), sigma (the 
            circular standard deviation), kappa, sigma_lo (alpha-level quantile 
            for sigma) and kappa_hi, the (1-alpha)-level quantile for kappa. 
                
        See also
        --------

        pyleoclim.core.series.Series.wavelet_coherence : Wavelet coherence

        pyleoclim.core.scalograms.Scalogram : Scalogram object

        pyleoclim.core.scalograms.MultipleScalogram : Multiple Scalogram object

        pyleoclim.core.coherence.Coherence.plot : plotting method for Coherence objects
        
        pyleoclime.utils.wavelet.angle_sig : significance of phase angle statistics
            
        pyleoclim.utils.wavelet.angle_stats: phase angle statistics
    
        
        References
        ----------
        
        [1] Grinsted, A., J. C. Moore, and S. Jevrejeva (2004), Application of the cross
        wavelet transform and wavelet coherence to geophysical time series, 
        Nonlinear Processes in Geophysics, 11, 561–566.
        
        [2] Huber, R., Dutra, L. V., & da Costa Freitas, C. (2001).
        SAR interferogram phase filtering based on the Von Mises distribution. 
        In IGARSS 2001. Scanning the Present and Resolving the Future. 
        Proceedings. IEEE 2001 International Geoscience and Remote Sensing Symposium 
        (Cat. No. 01CH37217) (Vol. 6, pp. 2816-2818). IEEE.  
        
        [3] Farge, M. and Schneider, K. (2006): Wavelets: application to turbulence
        Encyclopedia of Mathematical Physics (Eds. J.-P. Françoise, G. Naber and T.S. Tsun)
        pp 408-420. 

        Examples
        --------

        Calculate the phase angle between NINO3 and All India Rainfall at 5y scales:

        .. ipython:: python
            :okwarning:
            :okexcept:

            import pyleoclim as pyleo
            import pandas as pd
            import numpy as np
            data = pd.read_csv('https://raw.githubusercontent.com/LinkedEarth/Pyleoclim_util/Development/example_data/wtc_test_data_nino_even.csv')
            time = data['t'].values
            air = data['air'].values
            nino = data['nino'].values
            ts_air = pyleo.Series(time=time, value=air, time_name='Year (CE)',
                                  label='All India Rainfall', value_name='AIR (mm/month)')
            ts_nino = pyleo.Series(time=time, value=nino, time_name='Year (CE)',
                                   label='NINO3', value_name='NINO3 (K)')

            coh = ts_air.wavelet_coherence(ts_nino)
            coh.phase_stats(scales=5)
        
        One may also obtain phase angle statistics over an interval, like the 2-8y ENSO band: 
            
         .. ipython:: python
             :okwarning:
             :okexcept:
             
             phase = coh.phase_stats(scales=[2,8])
             print("The mean angle is {:4.2f}°".format(phase.mean_angle/np.pi*180))
             print(phase)
            
        From this example, one diagnoses a strong anti-phased relationship in the ENSO band,
        with high von Mises concentration (kappa ~ 3.35 >> kappa_hi) and low circular
        dispersion (sigma ~ 0.6 << sigma_lo). This would be strong evidence of a consistent
        anti-phasing between NINO3 and AIR at those scales.       
        '''
        scales = np.array(scales)
        
        if scales.max() > self.scale.max():
            warnings.warn("Requested scale exceeds largest scale in object. Truncating to "+str(self.scale.max()))
        
        if scales.size == 1:
            scale_idx = np.argmin(np.abs(self.scale - scales))
            res = waveutils.angle_sig(self.phase[:,scale_idx],nMC=number,level=level)
        elif  scales.size == 2:
            idx_lo = np.argmin(np.abs(self.scale - scales.min()))
            idx_hi = np.argmin(np.abs(self.scale - scales.max()))
            if (idx_hi >= idx_lo):
                raise ValueError("Insufficiently spaced scales. Please pick a single one, or a wider interval")
            else: # average phase over those scales
                nt, ns = self.phase.shape
                phase = np.empty((nt)) 
                for i in range(nt):
                    phase[i], _, _ = waveutils.angle_stats(self.phase[i,idx_hi:idx_lo]) 
                res = waveutils.angle_sig(phase,nMC=number,level=level) # assess significance 
            
        return res


''' Tests for pyleoclim.core.ui.PSD

Naming rules:
1. class: Test{filename}{Class}{method} with appropriate camel case
2. function: test_{method}_t{test_id}

Notes on how to test:
0. Make sure [pytest](https://docs.pytest.org) has been installed: `pip install pytest`
1. execute `pytest {directory_path}` in terminal to perform all tests in all testing files inside the specified directory
2. execute `pytest {file_path}` in terminal to perform all tests in the specified file
3. execute `pytest {file_path}::{TestClass}::{test_method}` in terminal to perform a specific test class/method inside the specified file
4. after `pip install pytest-xdist`, one may execute "pytest -n 4" to test in parallel with number of workers specified by `-n`
5. for more details, see https://docs.pytest.org/en/stable/usage.html
'''
import numpy as np
import pytest
import pyleoclim as pyleo

def gen_ts(model='colored_noise',alpha=1, nt=100, f0=None, m=None, seed=None):
    'wrapper for gen_ts in pyleoclim'
    
    t,v = pyleo.utils.gen_ts(model=model,alpha=alpha, nt=nt, f0=f0, m=m, seed=seed)
    ts=pyleo.Series(t,v)
    return ts

# a collection of useful functions

def gen_normal(loc=0, scale=1, nt=100):
    ''' Generate random data with a Gaussian distribution
    '''
    t = np.arange(nt)
    v = np.random.normal(loc=loc, scale=scale, size=nt)
    ts = pyleo.Series(t,v)
    return ts


# Tests below
class TestUiPsdPlot:
    ''' Tests for PSD.plot()
    '''

    def test_plot_t0(self):
        ''' Test PSD.plot() with default parameters
        '''
        ts = gen_ts(nt=500)
        psd = ts.spectral(method='mtm')
        fig, ax = psd.plot()
        pyleo.closefig(fig)

class TestUiPsdSignifTest:
    ''' Tests for PSD.signif_test()
    '''

    def test_signif_test_t0(self):
        ''' Test PSD.signif_test() with default parameters
        '''
        ts = gen_ts(nt=500)
        psd = ts.spectral(method='mtm')
        psd_signif = psd.signif_test(number=10)
        fig, ax = psd_signif.plot()
        pyleo.closefig(fig)


''' Tests for pyleoclim.core.ui.Scalogram

Naming rules:
1. class: Test{filename}{Class}{method} with appropriate camel case
2. function: test_{method}_t{test_id}

Notes on how to test:
0. Make sure [pytest](https://docs.pytest.org) has been installed: `pip install pytest`
1. execute `pytest {directory_path}` in terminal to perform all tests in all testing files inside the specified directory
2. execute `pytest {file_path}` in terminal to perform all tests in the specified file
3. execute `pytest {file_path}::{TestClass}::{test_method}` in terminal to perform a specific test class/method inside the specified file
4. after `pip install pytest-xdist`, one may execute "pytest -n 4" to test in parallel with number of workers specified by `-n`
5. for more details, see https://docs.pytest.org/en/stable/usage.html
'''

import pytest
import pyleoclim as pyleo

def gen_ts(model='colored_noise',alpha=1, nt=100, f0=None, m=None, seed=None):
    'wrapper for gen_ts in pyleoclim'
    
    t,v = pyleo.utils.gen_ts(model=model,alpha=alpha, nt=nt, f0=f0, m=m, seed=seed)
    ts=pyleo.Series(t,v)
    
    return ts

# Tests below
class TestUiScalogramSignifTest:
    ''' Tests for Scalogram.signif_test()
    '''

    @pytest.mark.parametrize('wave_method',['wwz','cwt'])
    def test_signif_test_t0(self, wave_method):
        ''' Test scalogram.signif_test() with default parameters
        '''
        ts = gen_ts(model='colored_noise',nt=500)
        scal = ts.wavelet(method=wave_method)
        scal_signif = scal.signif_test(number=5, qs = [0.8, 0.9, .95])
        fig,ax = scal_signif.plot(signif_thresh=0.99)
        pyleo.closefig(fig)
        

    @pytest.mark.parametrize('ar1_method',['ar1asym', 'ar1sim'])
    def test_signif_test_t1(self,ar1_method):
        ''' Test scalogram.signif_test() with default parameters
        '''
        ts = gen_ts(model='colored_noise',nt=500)
        scal = ts.wavelet(method='cwt')
        scal_signif = scal.signif_test(method=ar1_method,number=1)



# -*- coding: utf-8 -*-
"""
Created on Wed May 11 10:42:34 2016

@author: deborahkhider

License agreement - GNU GENERAL PUBLIC LICENSE v3
https://github.com/LinkedEarth/Pyleoclim_util/blob/master/license

"""
#Import pyleoclim modules

import pyleoclim.utils as utils
from .core import *

from .utils.plotting import *
set_style(style='journal', font_scale=1.4)

# get the version
from importlib.metadata import version
__version__ = version('pyleoclim')



import os
import sys
import io

from setuptools import setup, find_packages


version = '0.10.0'

# Read the readme file contents into variable
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name='pyleoclim',
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    version=version,
    license='GPL-3.0 License',
    description='A Python package for paleoclimate data analysis',
    long_description=read("README.md"),
    long_description_content_type = 'text/markdown',
    author='Deborah Khider, Feng Zhu, Julien Emile-Geay, Jun Hu, Myron Kwan, Pratheek Athreya, Alexander James, Daniel Garijo',
    author_email='linkedearth@gmail.com',
    url='https://github.com/LinkedEarth/Pyleoclim_util/pyleoclim',
    download_url='https://github.com/LinkedEarth/Pyleoclim_util/tarball/'+version,
    keywords=['Paleoclimate, Data Analysis, LiPD'],
    classifiers=[],
    install_requires=[
        "LiPD==0.2.8.8",
        "pandas>=1.3.0",
        "numpy<=1.24.0",
        "matplotlib>=3.6.0",
        "scipy>=1.9.1",
        "statsmodels>=0.13.2",
        "seaborn>=0.12.0",
        "scikit-learn>=0.24.2",
        "pathos>=0.2.8",
        "tqdm>=4.61.2",
        "tftb>=0.1.3",
        "pyhht>=0.1.0",
        "wget>=3.2",
        "numba>=0.56",
        "nitime>=0.9",
        "tabulate>=0.8.9",
        "Unidecode>=1.1.1",
    ],
    python_requires=">=3.8.0"
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Utilities for Liang and Granger causality analysis
"""

__all__ = [
    'liang_causality',
    'granger_causality'
]

import numpy as np
from statsmodels.tsa.stattools import grangercausalitytests
from tqdm import tqdm
from .tsmodel import ar1_fit_evenly
from .correlation import sm_ar1_sim, phaseran
from scipy.stats.mstats import mquantiles

#-------
#Main functions
#--------
def granger_causality(y1, y2, maxlag=1,addconst=True,verbose=True):
    '''Granger causality tests
    
    Four tests for the Granger non-causality of 2 time series.

    All four tests give similar results. params_ftest and ssr_ftest are equivalent based on F test which is identical to lmtest:grangertest in R.
    
    Wrapper for the functions described in statsmodels (https://www.statsmodels.org/stable/generated/statsmodels.tsa.stattools.grangercausalitytests.html)

    Parameters
    ----------

    y1, y2: array
        vectors of (real) numbers with identical length, no NaNs allowed
    maxlag : int or int iterable, optional
        If an integer, computes the test for all lags up to maxlag. If an iterable, computes the tests only for the lags in maxlag.
    addconst : bool, optional
        Include a constant in the model.
    verbose : bool, optional
        Print results

    Returns
    -------

    dict
        All test results, dictionary keys are the number of lags. For each lag the values are a tuple, with the first element a dictionary with test statistic,
        pvalues, degrees of freedom, the second element are the OLS estimation results for the restricted model, the unrestricted model and the restriction (contrast)
        matrix for the parameter f_test.
        
    Notes
    -----

    The null hypothesis for Granger causality tests is that y2, does NOT Granger cause y1. Granger causality means that past values of y2 have a statistically significant effect on the current value of y1, taking past values of y1 into account as regressors. We reject the null hypothesis that y2 does not Granger cause y1 if the p-values are below a desired threshold (e.g. 0.05).

    The null hypothesis for all four test is that the coefficients corresponding to past values of the second time series are zero.

    ‘params_ftest’, ‘ssr_ftest’ are based on the F distribution

    ‘ssr_chi2test’, ‘lrtest’ are based on the chi-square distribution
    
    See also
    --------

    pyleoclim.utils.causality.liang_causality : information flow estimated using the Liang algorithm
    
    pyleoclim.utils.causality.signif_isopersist : significance test with AR(1) with same persistence
    
    pyleoclim.utils.causality.signif_isospec : significance test with surrogates with randomized phases
    
    References
    ----------

    Granger, C. W. J. (1969). Investigating causal relations by econometric models and cross-spectral methods. Econometrica, 37(3), 424-438. 
    
    Granger, C. W. J. (1980). Testing for causality: A personal viewpoont. Journal of Economic Dynamics and Control, 2, 329-352. 
    
    Granger, C. W. J. (1988). Some recent development in a concept of causality. Journal of Econometrics, 39(1-2), 199-211. 
    
    '''

    if len(y1)!=len(y2):
        raise ValueError('Timeseries must be of same length')

    x=np.array([y1,y2]).T
    res = grangercausalitytests(x,maxlag=maxlag,addconst=addconst,verbose=verbose)
    return res

def liang_causality(y1, y2, npt=1, signif_test='isospec', nsim=1000,
                    qs=[0.005, 0.025, 0.05, 0.95, 0.975, 0.995]):
    '''Liang-Kleeman information flow
    
    Estimate the Liang information transfer from series y2 to series y1 with 
    significance estimates using either an AR(1) tests with series with the same
    persistence or surrogates with randomized phases.

    Parameters
    ----------

    y1, y2 : array
        vectors of (real) numbers with identical length, no NaNs allowed

    npt : int >=1
        time advance in performing Euler forward differencing,
        e.g., 1, 2. Unless the series are generated with a highly chaotic deterministic system,
        npt=1 should be used
    
    signif_test : str; {'isopersist', 'isospec'}
        the method for significance test
        see signif_isospec and signif_isopersist for details. 
        
    nsim : int
        the number of AR(1) surrogates for significance test
        
    qs : list
        the quantiles for significance test

    Returns
    -------

    res : dict
        A dictionary of results including:
            T21 : float
                information flow from y2 to y1 (Note: not y1 -> y2!)
            tau21 : float
                the standardized information flow from y2 to y1
            Z : float
                the total information flow from y2 to y1
            dH1_star : float
                dH*/dt (Liang, 2016)
            dH1_noise : float
            signif_qs : 
                the quantiles for significance test
            T21_noise : list
                the quantiles of the information flow from noise2 to noise1 for significance testing
            tau21_noise : list
                the quantiles of the standardized information flow from noise2 to noise1 for significance testing
    
    See also
    --------

    pyleoclim.utils.causality.liang : information flow estimated using the Liang algorithm
    
    pyleoclim.utils.causality.granger_causality : information flow estimated using the Granger algorithm
    
    pyleoclim.utils.causality.signif_isopersist : significance test with AR(1) with same persistence
    
    pyleoclim.utils.causality.causality.signif_isospec : significance test with surrogates with randomized phases
    
    References
    ----------

    Liang, X.S. (2013) The Liang-Kleeman Information Flow: Theory and Applications. Entropy, 15, 327-360, doi:10.3390/e15010327
    
    Liang, X.S. (2014) Unraveling the cause-effect relation between timeseries. Physical review, E 90, 052150
    
    Liang, X.S. (2015) Normalizing the causality between time series. Physical review, E 92, 022126
    
    Liang, X.S. (2016) Information flow and causality as rigorous notions ab initio. Physical review, E 94, 052201

    '''

    dt=1
    nm = np.size(y1)

    grad1 = (y1[0+npt:] - y1[0:-npt]) / (npt)
    grad2 = (y2[0+npt:] - y2[0:-npt]) / (npt)

    y1 = y1[:-npt]
    y2 = y2[:-npt]

    N = nm - npt
    C = np.cov(y1, y2)
    detC = np.linalg.det(C)

    dC = np.ndarray((2, 2))
    dC[0, 0] = np.sum((y1-np.mean(y1))*(grad1-np.mean(grad1)))
    dC[0, 1] = np.sum((y1-np.mean(y1))*(grad2-np.mean(grad2)))
    dC[1, 0] = np.sum((y2-np.mean(y2))*(grad1-np.mean(grad1)))
    dC[1, 1] = np.sum((y2-np.mean(y2))*(grad2-np.mean(grad2)))

    dC /= N-1

    a11 = C[1, 1]*dC[0, 0] - C[0, 1]*dC[1, 0]
    a12 = -C[0, 1]*dC[0, 0] + C[0, 0]*dC[1, 0]

    a11 /= detC
    a12 /= detC

    f1 = np.mean(grad1) - a11*np.mean(y1) - a12*np.mean(y2)
    R1 = grad1 - (f1 + a11*y1 + a12*y2)
    Q1 = np.sum(R1*R1)
    b1 = np.sqrt(Q1*dt/N)

    NI = np.ndarray((4, 4))
    NI[0, 0] = N*dt/b1**2
    NI[1, 1] = dt/b1**2*np.sum(y1*y1)
    NI[2, 2] = dt/b1**2*np.sum(y2*y2)
    NI[3, 3] = 3*dt/b1**4*np.sum(R1*R1) - N/b1**2
    NI[0, 1] = dt/b1**2*np.sum(y1)
    NI[0, 2] = dt/b1**2*np.sum(y2)
    NI[0, 3] = 2*dt/b1**3*np.sum(R1)
    NI[1, 2] = dt/b1**2*np.sum(y1*y2)
    NI[1, 3] = 2*dt/b1**3*np.sum(R1*y1)
    NI[2, 3] = 2*dt/b1**3*np.sum(R1*y2)

    NI[1, 0] = NI[0, 1]
    NI[2, 0] = NI[0, 2]
    NI[2, 1] = NI[1, 2]
    NI[3, 0] = NI[0, 3]
    NI[3, 1] = NI[1, 3]
    NI[3, 2] = NI[2, 3]

    invNI = np.linalg.pinv(NI)
    var_a12 = invNI[2, 2]
    T21 = C[0, 1]/C[0, 0] * (-C[1, 0]*dC[0, 0] + C[0, 0]*dC[1, 0]) / detC
    var_T21 = (C[0, 1]/C[0, 0])**2 * var_a12

    dH1_star= a11
    dH1_noise = b1**2 / (2*C[0, 0])

    Z = np.abs(T21) + np.abs(dH1_star) + np.abs(dH1_noise)

    tau21 = T21 / Z
    dH1_star = dH1_star / Z
    dH1_noise = dH1_noise / Z
    
    signif_test_func = {
            'isopersist': signif_isopersist,
            'isospec': signif_isospec,
        }
    
    signif_dict = signif_test_func[signif_test](y1, y2, method='liang', nsim=nsim, qs=qs, npt=npt)
    T21_noise_qs = signif_dict['T21_noise_qs']
    tau21_noise_qs = signif_dict['tau21_noise_qs']

    res = {
        'T21': T21,
        'tau21': tau21,
        'Z': Z,
        'dH1_star': dH1_star,
        'dH1_noise': dH1_noise,
        'signif_qs' : qs,
        'T21_noise' : T21_noise_qs,
        'tau21_noise' : tau21_noise_qs
    }

    return res

def liang(y1, y2, npt=1):
    '''
    Estimate the Liang information transfer from series y2 to series y1 

    Parameters
    ----------

    y1, y2 : array
        Vectors of (real) numbers with identical length, no NaNs allowed

    npt : int  >=1
        Time advance in performing Euler forward differencing,
        e.g., 1, 2. Unless the series are generated with a highly chaotic deterministic system,
        npt=1 should be used

    Returns
    -------

    res : dict
        A dictionary of results including:
            T21 : float
                information flow from y2 to y1 (Note: not y1 -> y2!)
            tau21 : float
                the standardized information flow from y2 to y1
            Z : float
                the total information flow from y2 to y1
            dH1_star : float
                dH*/dt (Liang, 2016)
            dH1_noise : float
            
    See also
    --------

    pyleoclim.utils.causality.liang_causality : information flow estimated using the Liang algorithm
    pyleoclim.utils.causality.granger_causality : information flow estimated using the Granger algorithm    
    pyleoclim.utils.causality.signif_isopersist : significance test with AR(1) with same persistence
    pyleoclim.utils.causality.signif_isospec : significance test with surrogates with randomized phases
    
    References
    ----------

    Liang, X.S. (2013) The Liang-Kleeman Information Flow: Theory and
            Applications. Entropy, 15, 327-360, doi:10.3390/e15010327
    
    Liang, X.S. (2014) Unraveling the cause-effect relation between timeseries.
        Physical review, E 90, 052150
    
    Liang, X.S. (2015) Normalizing the causality between time series.
        Physical review, E 92, 022126
    
    Liang, X.S. (2016) Information flow and causality as rigorous notions ab initio.
        Physical review, E 94, 052201

    '''
    dt=1
    nm = np.size(y1)

    grad1 = (y1[0+npt:] - y1[0:-npt]) / (npt)
    grad2 = (y2[0+npt:] - y2[0:-npt]) / (npt)

    y1 = y1[:-npt]
    y2 = y2[:-npt]

    N = nm - npt
    C = np.cov(y1, y2)
    detC = np.linalg.det(C)

    dC = np.ndarray((2, 2))
    dC[0, 0] = np.sum((y1-np.mean(y1))*(grad1-np.mean(grad1)))
    dC[0, 1] = np.sum((y1-np.mean(y1))*(grad2-np.mean(grad2)))
    dC[1, 0] = np.sum((y2-np.mean(y2))*(grad1-np.mean(grad1)))
    dC[1, 1] = np.sum((y2-np.mean(y2))*(grad2-np.mean(grad2)))

    dC /= N-1

    a11 = C[1, 1]*dC[0, 0] - C[0, 1]*dC[1, 0]
    a12 = -C[0, 1]*dC[0, 0] + C[0, 0]*dC[1, 0]

    a11 /= detC
    a12 /= detC

    f1 = np.mean(grad1) - a11*np.mean(y1) - a12*np.mean(y2)
    R1 = grad1 - (f1 + a11*y1 + a12*y2)
    Q1 = np.sum(R1*R1)
    b1 = np.sqrt(Q1*dt/N)

    NI = np.ndarray((4, 4))
    NI[0, 0] = N*dt/b1**2
    NI[1, 1] = dt/b1**2*np.sum(y1*y1)
    NI[2, 2] = dt/b1**2*np.sum(y2*y2)
    NI[3, 3] = 3*dt/b1**4*np.sum(R1*R1) - N/b1**2
    NI[0, 1] = dt/b1**2*np.sum(y1)
    NI[0, 2] = dt/b1**2*np.sum(y2)
    NI[0, 3] = 2*dt/b1**3*np.sum(R1)
    NI[1, 2] = dt/b1**2*np.sum(y1*y2)
    NI[1, 3] = 2*dt/b1**3*np.sum(R1*y1)
    NI[2, 3] = 2*dt/b1**3*np.sum(R1*y2)

    NI[1, 0] = NI[0, 1]
    NI[2, 0] = NI[0, 2]
    NI[2, 1] = NI[1, 2]
    NI[3, 0] = NI[0, 3]
    NI[3, 1] = NI[1, 3]
    NI[3, 2] = NI[2, 3]

    invNI = np.linalg.pinv(NI)
    var_a12 = invNI[2, 2]
    T21 = C[0, 1]/C[0, 0] * (-C[1, 0]*dC[0, 0] + C[0, 0]*dC[1, 0]) / detC
    var_T21 = (C[0, 1]/C[0, 0])**2 * var_a12

    dH1_star= a11
    dH1_noise = b1**2 / (2*C[0, 0])

    Z = np.abs(T21) + np.abs(dH1_star) + np.abs(dH1_noise)

    tau21 = T21 / Z
    dH1_star = dH1_star / Z
    dH1_noise = dH1_noise / Z

    res = {
        'T21': T21,
        'tau21': tau21,
        'Z': Z,
        'dH1_star': dH1_star,
        'dH1_noise': dH1_noise,
    }

    return res

def signif_isopersist(y1, y2, method,
                      nsim=1000, qs=[0.005, 0.025, 0.05, 0.95, 0.975, 0.995],
                      **kwargs):
    ''' significance test with AR(1) with same persistence

    Parameters
    ----------

    y1, y2 : array
        vectors of (real) numbers with identical length, no NaNs allowed

    method : str; {'liang'}
        estimates for the Liang method

    nsim : int
        the number of AR(1) surrogates for significance test

    qs : list
        the quantiles for significance test

    Returns
    -------

    res_dict : dict
        A dictionary with the following information:
          T21_noise_qs : list
            the quantiles of the information flow from noise2 to noise1 for significance testing
          tau21_noise_qs : list
            the quantiles of the standardized information flow from noise2 to noise1 for significance testing

    See also
    --------

    pyleoclim.utils.causality.liang_causality : information flow estimated using the Liang algorithm
    pyleoclim.utils.causality.granger_causality : information flow estimated using the Granger algorithm    
    pyleoclim.utils.causality.signif_isospec : significance test with surrogates with randomized phases
    
    '''
    g1 = ar1_fit_evenly(y1)
    g2 = ar1_fit_evenly(y2)
    sig1 = np.std(y1)
    sig2 = np.std(y2)
    n = np.size(y1)
    noise1 = sm_ar1_sim(n, nsim, g1, sig1)
    noise2 = sm_ar1_sim(n, nsim, g2, sig2)

    if method == 'liang':
        npt = kwargs['npt'] if 'npt' in kwargs else 1
        T21_noise = []
        tau21_noise = []
        for i in tqdm(range(nsim), desc='Calculating causality between surrogates'):
            res_noise = liang(noise1[:, i], noise2[:, i], npt=npt)
            tau21_noise.append(res_noise['tau21'])
            T21_noise.append(res_noise['T21'])
        tau21_noise = np.array(tau21_noise)
        T21_noise = np.array(T21_noise)
        tau21_noise_qs = mquantiles(tau21_noise, qs)
        T21_noise_qs = mquantiles(T21_noise, qs)

        res_dict = {
            'tau21_noise_qs': tau21_noise_qs,
            'T21_noise_qs': T21_noise_qs,
        }
    #TODO add granger method
    else:
        raise KeyError(f'{method} is not a valid method')

    return res_dict

def signif_isospec(y1, y2, method,
                   nsim=1000, qs=[0.005, 0.025, 0.05, 0.95, 0.975, 0.995],
                   **kwargs):
    ''' significance test with surrogates with randomized phases

    Parameters
    ----------

    y1, y2 : array
        vectors of (real) numbers with identical length, no NaNs allowed
    method : str; {'liang'}
        estimates for the Liang method
    nsim : int
        the number of surrogates for significance test
    qs : list
        the quantiles for significance test
    kwargs : dict
        keyword arguments for the causality method (e.g. npt for Liang-Kleeman)

    Returns
    -------

    res_dict : dict
        A dictionary with the following information:
          T21_noise_qs : list
                        the quantiles of the information flow from noise2 to noise1 for significance testing
          tau21_noise_qs : list
                          the quantiles of the standardized information flow from noise2 to noise1 for significance testing
    
    See also
    --------

    pyleoclim.utils.causality.liang_causality : information flow estimated using the Liang algorithm
    pyleoclim.utils.causality.granger_causality : information flow estimated using the Granger algorithm    
    pyleoclim.utils.causality.signif_isopersist : significance test with AR(1) with same persistence
    
    '''
    
    noise1 = phaseran(y1, nsim)
    noise2 = phaseran(y2, nsim)

    if method == 'liang':
        npt = kwargs['npt'] if 'npt' in kwargs else 1
        T21_noise = []
        tau21_noise = []
        for i in tqdm(range(nsim), desc='Calculating causality between surrogates'):
            res_noise = liang(noise1[:, i], noise2[:, i], npt=npt)
            tau21_noise.append(res_noise['tau21'])
            T21_noise.append(res_noise['T21'])
        tau21_noise = np.array(tau21_noise)
        T21_noise = np.array(T21_noise)
        tau21_noise_qs = mquantiles(tau21_noise, qs)
        T21_noise_qs = mquantiles(T21_noise, qs)

        res_dict = {
            'tau21_noise_qs': tau21_noise_qs,
            'T21_noise_qs': T21_noise_qs,
        }
    #TODO Recode with Surrogate class
    else:
        raise KeyError(f'{method} is not a valid method')

    return res_dict



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
LipdSeries are (you guessed it), Series objects that are created from LiPD objects. As a subclass of Series, they inherit all its methods.
When created, LiPDSeries automatically instantiates the time, value and other parameters from what’s in the lipd file.

"""

from ..utils import plotting, mapping, lipdutils
from ..core.series import Series
from ..core.ensembleseries import EnsembleSeries

import matplotlib.pyplot as plt
import numpy as np
from copy import deepcopy

from matplotlib import gridspec
import cartopy.crs as ccrs
import cartopy.feature as cfeature
import warnings
import os
import lipd as lpd


class LipdSeries(Series):
    '''LipdSeries are (you guessed it), Series objects that are created from LiPD objects. As a subclass of Series, they inherit all its methods.
    When created, LiPDSeries automatically instantiates the time, value and other parameters from what’s in the lipd file.
    These objects can be obtained from a LiPD file/object either through Pyleoclim or the LiPD utilities.
    If multiple objects (i.e., a list) are given, then the user will be prompted to choose one timeseries.

    Returns
    -------
    object : pyleoclim.LipdSeries

    See also
    --------

    pyleoclim.core.lipd.Lipd : Creates a Lipd object from LiPD Files

    pyleoclim.core.series.Series : Creates pyleoclim Series object

    pyleoclim.core.multipleseries.MultipleSeries : a collection of multiple Series objects

    Examples
    --------

    In this example, we will import a LiPD file and explore the various options to create a series object.

    First, let's look at the Lipd.to_tso option. This method is attractive because the object is a list of dictionaries that are easily explored in Python.

    .. ipython:: python
        :okwarning:
        :okexcept:

        import pyleoclim as pyleo
        url = 'http://wiki.linked.earth/wiki/index.php/Special:WTLiPD?op=export&lipdid=MD982176.Stott.2004'
        data = pyleo.Lipd(usr_path = url)
        ts_list = data.to_tso()
        # Print out the dataset name and the variable name
        for item in ts_list:
            print(item['dataSetName']+': '+item['paleoData_variableName'])
        # Load the sst data into a LipdSeries. Since Python indexing starts at zero, sst has index 5.
        ts = pyleo.LipdSeries(ts_list[5])

    If you attempt to pass the full list of series, Pyleoclim will prompt you to choose a series by printing out something similar as above.
    If you already now the number of the timeseries object you're interested in, then you should use the following:

    .. ipython:: python
        :okwarning:
        :okexcept:

        ts1 = data.to_LipdSeries(number=5)

    If number is not specified, Pyleoclim will prompt you for the number automatically.

    Sometimes, one may want to create a MultipleSeries object from a collection of LiPD files. In this case, we recommend using the following:

    .. ipython:: python
        :okwarning:
        :okexcept:

        ts_list = data.to_LipdSeriesList()
        # only keep the Mg/Ca and SST
        ts_list=ts_list[4:]
        #create a MultipleSeries object
        ms=pyleo.MultipleSeries(ts_list)

    '''

    def __init__(self, tso, clean_ts=True, verbose=False):
        if type(tso) is list:
            self.lipd_ts = lipdutils.getTs(tso)
        else:
            self.lipd_ts = tso

        self.plot_default = {'ice-other': ['#FFD600', 'h'],
                             'ice/rock': ['#FFD600', 'h'],
                             'coral': ['#FF8B00', 'o'],
                             'documents': ['k', 'p'],
                             'glacierice': ['#86CDFA', 'd'],
                             'hybrid': ['#00BEFF', '*'],
                             'lakesediment': ['#4169E0', 's'],
                             'marinesediment': ['#8A4513', 's'],
                             'sclerosponge': ['r', 'o'],
                             'speleothem': ['#FF1492', 'd'],
                             'wood': ['#32CC32', '^'],
                             'molluskshells': ['#FFD600', 'h'],
                             'peat': ['#2F4F4F', '*'],
                             'midden': ['#824E2B', 'o'],
                             'other': ['k', 'o']}

        try:
            time, label = lipdutils.checkTimeAxis(self.lipd_ts)
            if label == 'age':
                time_name = 'Age'
                if 'ageUnits' in self.lipd_ts.keys():
                    time_unit = self.lipd_ts['ageUnits']
                else:
                    time_unit = None
            elif label == 'year':
                time_name = 'Year'
                if 'yearUnits' in self.lipd_ts.keys():
                    time_unit = self.lipd_ts['yearUnits']
                else:
                    time_unit = None
            try:
                if self.lipd_ts['mode'] == 'paleoData':
                    value = np.array(self.lipd_ts['paleoData_values'], dtype='float64')
                    value_name = self.lipd_ts['paleoData_variableName']
                    if 'paleoData_units' in self.lipd_ts.keys():
                        value_unit = self.lipd_ts['paleoData_units']
                    else:
                        value_unit = None
                    label = self.lipd_ts['dataSetName']
                    super(LipdSeries, self).__init__(time=time, value=value, time_name=time_name,
                                                     time_unit=time_unit, value_name=value_name, value_unit=value_unit,
                                                     label=label, clean_ts=clean_ts, verbose=verbose)
                elif self.lipd_ts['mode'] == 'chronData':
                    value = np.array(self.lipd_ts['chronData_values'], dtype='float64')
                    value_name = self.lipd_ts['chronData_variableName']
                    if 'paleoData_units' in self.lipd_ts.keys():
                        value_unit = self.lipd_ts['chronData_units']
                    else:
                        value_unit = None
                    label = self.lipd_ts['dataSetName']
                    super(LipdSeries, self).__init__(time=time, value=value, time_name=time_name,
                                                     time_unit=time_unit, value_name=value_name, value_unit=value_unit,
                                                     label=label, clean_ts=clean_ts, verbose=verbose)
            except:
                raise ValueError("paleoData_values should contain floats")
        except:
            raise KeyError("No time information present")

    def copy(self):
        '''Copy the object

        Returns
        -------
        
        object : pyleoclim.LipdSeries
            New object with data copied from original

        '''
        return deepcopy(self)

    def chronEnsembleToPaleo(self, D, number=None, chronNumber=None, modelNumber=None, tableNumber=None):
        '''Fetch chron ensembles from a Lipd object and return the ensemble as MultipleSeries

        Parameters
        ----------

        D : a LiPD object

        number: int, optional

            The number of ensemble members to store. Default is None, which corresponds to all present

        chronNumber: int, optional

            The chron object number. The default is None.

        modelNumber : int, optional

            Age model number. The default is None.

        tableNumber : int, optional

            Table number. The default is None.

        Raises
        ------

        ValueError

        Returns
        -------

        ens : EnsembleSeries
            An EnsembleSeries object with each series representing a possible realization of the age model

        See also
        --------

        pyleoclim.core.ensembleseries.EnsembleSeries : An EnsembleSeries object with each series representing a possible realization of the age model

        pyleoclim.utils.lipdutils.mapAgeEnsembleToPaleoData : Map the depth for the ensemble age values to the paleo depth

        '''
        # get the corresponding LiPD
        dataSetName = self.lipd_ts['dataSetName']
        if type(D) is dict:
            try:
                lipd = D[dataSetName]
            except:
                lipd = D
        else:
            a = D.extract(dataSetName)
            lipd = a.__dict__['lipd']
        # Look for the ensemble and get values
        cwd = os.getcwd()
        csv_dict = lpd.getCsv(lipd)
        os.chdir(cwd)
        chron, paleo = lipdutils.isEnsemble(csv_dict)
        if len(chron) == 0:
            raise ValueError("No ChronMeasurementTables available")
        elif len(chron) > 1:
            if chronNumber == None or modelNumber == None or tableNumber == None:
                csvName = lipdutils.whichEnsemble(chron)
            else:
                str0 = 'chron' + str(chronNumber)
                str1 = 'model' + str(modelNumber)
                str2 = 'ensemble' + str(tableNumber)
                for item in chron:
                    if str0 in item and str1 in item and str2 in item:
                        csvName = item
            depth, ensembleValues = lipdutils.getEnsemble(csv_dict, csvName)
        else:
            depth, ensembleValues = lipdutils.getEnsemble(csv_dict, chron[0])
        # make sure it's sorted
        sort_ind = np.argsort(depth)
        depth = list(np.array(depth)[sort_ind])
        ensembleValues = ensembleValues[sort_ind, :]

        if number is not None:
            if number > np.shape(ensembleValues)[1]:
                warnings.warn(
                    'Selected number of ensemble members is greater than number of members in the ensemble table; passing')
                pass
            else:
                ensembleValues = ensembleValues[:, 0:number]

        # Map to paleovalues
        key = []
        for item in self.lipd_ts.keys():
            if 'depth' in item and 'Units' not in item:
                key.append(item)
        key = key[0]
        ds = np.array(self.lipd_ts[key], dtype='float64')
        if 'paleoData_values' in self.lipd_ts.keys():
            ys = np.array(self.lipd_ts['paleoData_values'], dtype='float64')
        elif 'chronData_values' in self.lipd_ts.keys():
            ys = np.array(self.lipd_ts['chronData_values'], dtype='float64')
        else:
            raise KeyError('no y-axis values available')
        # Remove NaNs
        ys_tmp = np.copy(ys)
        ds = ds[~np.isnan(ys_tmp)]
        sort_ind2 = np.argsort(ds)
        ds = np.array(ds[sort_ind2])
        ys = np.array(ys[sort_ind2])
        ensembleValuestoPaleo = lipdutils.mapAgeEnsembleToPaleoData(ensembleValues, depth, ds)
        # create multipleseries
        s_list = []
        for i, s in enumerate(ensembleValuestoPaleo.T):
            s_tmp = Series(time=s, value=ys,
                           verbose=i == 0,
                           clean_ts=False,
                           value_name=self.value_name,
                           value_unit=self.value_unit,
                           time_name=self.time_name,
                           time_unit=self.time_unit)
            s_list.append(s_tmp)

        ens = EnsembleSeries(series_list=s_list)

        return ens

    def map(self, projection='Orthographic', proj_default=True,
            background=True, borders=False, rivers=False, lakes=False,
            figsize=None, ax=None, marker=None, color=None,
            markersize=None, scatter_kwargs=None,
            legend=True, lgd_kwargs=None, savefig_settings=None):
        '''Map the location of the record

        Parameters
        ----------
        
        projection : str, optional

            The projection to use. The default is 'Robinson'.

        proj_default : bool; {True, False}, optional

            Whether to use the Pyleoclim defaults for each projection type. The default is True.

        background :  bool; {True, False}, optional

            Whether to use a background. The default is True.

        borders :  bool; {True, False}, optional

            Draw borders. The default is False.

        rivers :  bool; {True, False}, optional

            Draw rivers. The default is False.

        lakes :  bool; {True, False}, optional

            Draw lakes. The default is False.

        figsize : list or tuple, optional

            The size of the figure. The default is None.

        ax : matplotlib.ax, optional

            The matplotlib axis onto which to return the map. The default is None.

        marker : str, optional

            The marker type for each archive. The default is None. Uses plot_default

        color : str, optional

            Color for each archive. The default is None. Uses plot_default

        markersize : float, optional

            Size of the marker. The default is None.

        scatter_kwargs : dict, optional

            Parameters for the scatter plot. The default is None.

        legend :  bool; {True, False}, optional

            Whether to plot the legend. The default is True.

        lgd_kwargs : dict, optional

            Arguments for the legend. The default is None.

        savefig_settings : dict, optional

            the dictionary of arguments for plt.savefig(); some notes below:
            - "path" must be specified; it can be any existed or non-existed path,
              with or without a suffix; if the suffix is not given in "path", it will follow "format"
            - "format" can be one of {"pdf", "eps", "png", "ps"}. The default is None.

        Returns
        -------

        res : fig,ax

        See also
        --------

        pyleoclim.utils.mapping.map : Underlying mapping function for Pyleoclim

        Examples
        --------

        .. ipython:: python
            :okwarning:
            :okexcept:

            import pyleoclim as pyleo
            url = 'http://wiki.linked.earth/wiki/index.php/Special:WTLiPD?op=export&lipdid=MD982176.Stott.2004'
            data = pyleo.Lipd(usr_path = url)
            ts = data.to_LipdSeries(number=5)
            @savefig mapone.png
            fig, ax = ts.map()
            pyleo.closefig(fig)

        '''
        scatter_kwargs = {} if scatter_kwargs is None else scatter_kwargs.copy()
        # get the information from the timeseries
        lat = [self.lipd_ts['geo_meanLat']]
        lon = [self.lipd_ts['geo_meanLon']]

        if 'archiveType' in self.lipd_ts.keys():
            archiveType = lipdutils.LipdToOntology(self.lipd_ts['archiveType']).lower().replace(" ", "")
        else:
            archiveType = 'other'

        # make sure criteria is in the plot_default list
        if archiveType not in self.plot_default.keys():
            archiveType = 'other'

        if markersize is not None:
            scatter_kwargs.update({'s': markersize})

        if marker == None:
            marker = self.plot_default[archiveType][1]

        if color == None:
            color = self.plot_default[archiveType][0]

        if proj_default == True:
            proj1 = {'central_latitude': lat[0],
                     'central_longitude': lon[0]}
            proj2 = {'central_latitude': lat[0]}
            proj3 = {'central_longitude': lon[0]}

        archiveType = [archiveType]  # list so it will work with map
        marker = [marker]
        color = [color]

        if proj_default == True:

            try:
                res = mapping.map(lat=lat, lon=lon, criteria=archiveType,
                                  marker=marker, color=color,
                                  projection=projection, proj_default=proj1,
                                  background=background, borders=borders,
                                  rivers=rivers, lakes=lakes,
                                  figsize=figsize, ax=ax,
                                  scatter_kwargs=scatter_kwargs, legend=legend,
                                  lgd_kwargs=lgd_kwargs, savefig_settings=savefig_settings, )

            except:
                try:
                    res = mapping.map(lat=lat, lon=lon, criteria=archiveType,
                                      marker=marker, color=color,
                                      projection=projection, proj_default=proj3,
                                      background=background, borders=borders,
                                      rivers=rivers, lakes=lakes,
                                      figsize=figsize, ax=ax,
                                      scatter_kwargs=scatter_kwargs, legend=legend,
                                      lgd_kwargs=lgd_kwargs, savefig_settings=savefig_settings)
                except:
                    res = mapping.map(lat=lat, lon=lon, criteria=archiveType,
                                      marker=marker, color=color,
                                      projection=projection, proj_default=proj2,
                                      background=background, borders=borders,
                                      rivers=rivers, lakes=lakes,
                                      figsize=figsize, ax=ax,
                                      scatter_kwargs=scatter_kwargs, legend=legend,
                                      lgd_kwargs=lgd_kwargs, savefig_settings=savefig_settings)

        else:
            res = mapping.map(lat=lat, lon=lon, criteria=archiveType,
                              marker=marker, color=color,
                              projection=projection, proj_default=proj_default,
                              background=background, borders=borders,
                              rivers=rivers, lakes=lakes,
                              figsize=figsize, ax=ax,
                              scatter_kwargs=scatter_kwargs, legend=legend,
                              lgd_kwargs=lgd_kwargs, savefig_settings=savefig_settings)
        return res

    def getMetadata(self):

        """ Get the necessary metadata for the ensemble plots

        Parameters
        ----------

        timeseries : object

                    a specific timeseries object.

        Returns
        -------

        res : dict
                  A dictionary containing the following metadata:
                    archiveType,
                    Authors (if more than 2, replace by et al),
                    PublicationYear,
                    Publication DOI,
                    Variable Name,
                    Units,
                    Climate Interpretation,
                    Calibration Equation,
                    Calibration References,
                    Calibration Notes

        """

        # Get all the necessary information
        # Top level information
        if "archiveType" in self.lipd_ts.keys():
            archiveType = self.lipd_ts["archiveType"]
        else:
            archiveType = "NA"

        if "pub1_author" in self.lipd_ts.keys():
            authors = self.lipd_ts["pub1_author"]
        else:
            authors = "NA"

        # Truncate if more than two authors
        idx = [pos for pos, char in enumerate(authors) if char == ";"]
        if len(idx) > 2:
            authors = authors[0:idx[1] + 1] + "et al."

        if "pub1_year" in self.lipd_ts.keys():
            Year = str(self.lipd_ts["pub1_year"])
        else:
            Year = "NA"

        if "pub1_doi" in self.lipd_ts.keys():
            DOI = self.lipd_ts["pub1_doi"]
        else:
            DOI = "NA"

        if self.lipd_ts['mode'] == 'paleoData':
            prefix = 'paleo'
        else:
            prefix = 'chron'

        if prefix + "Data_InferredVariableType" in self.lipd_ts.keys():
            if type(self.lipd_ts[prefix + "Data_InferredVariableType"]) is list:
                Variable = self.lipd_ts[prefix + "Data_InferredVariableType"][0]
            else:
                Variable = self.lipd_ts[prefix + "Data_InferredVariableType"]
        elif prefix + "Data_ProxyObservationType" in self.lipd_ts.keys():
            if type(self.lipd_ts[prefix + "Data_ProxyObservationType"]) is list:
                Variable = self.lipd_ts[prefix + "Data_ProxyObservationType"][0]
            else:
                Variable = self.lipd_ts[prefix + "Data_ProxyObservationType"]
        else:
            Variable = self.lipd_ts[prefix + "Data_variableName"]

        if prefix + "Data_units" in self.lipd_ts.keys():
            units = self.lipd_ts[prefix + "Data_units"]
        else:
            units = "NA"

        # Climate interpretation information
        if prefix + "Data_interpretation" in self.lipd_ts.keys():
            interpretation = self.lipd_ts[prefix + "Data_interpretation"][0]
            if "name" in interpretation.keys():
                ClimateVar = interpretation["name"]
            elif "variable" in interpretation.keys():
                ClimateVar = interpretation["variable"]
            else:
                ClimateVar = "NA"
            if "detail" in interpretation.keys():
                Detail = interpretation["detail"]
            elif "variableDetail" in interpretation.keys():
                Detail = interpretation['variableDetail']
            else:
                Detail = "NA"
            if "scope" in interpretation.keys():
                Scope = interpretation['scope']
            else:
                Scope = "NA"
            if "seasonality" in interpretation.keys():
                Seasonality = interpretation["seasonality"]
            else:
                Seasonality = "NA"
            if "interpdirection" in interpretation.keys():
                Direction = interpretation["interpdirection"]
            else:
                Direction = "NA"
        else:
            ClimateVar = "NA"
            Detail = "NA"
            Scope = "NA"
            Seasonality = "NA"
            Direction = "NA"

        # Calibration information
        if prefix + "Data_calibration" in self.lipd_ts.keys():
            calibration = self.lipd_ts[prefix + 'Data_calibration'][0]
            if "equation" in calibration.keys():
                Calibration_equation = calibration["equation"]
            else:
                Calibration_equation = "NA"
            if "calibrationReferences" in calibration.keys():
                ref = calibration["calibrationReferences"]
                if "author" in ref.keys():
                    ref_author = ref["author"][0]  # get the first author
                else:
                    ref_author = "NA"
                if "publicationYear" in ref.keys():
                    ref_year = str(ref["publicationYear"])
                else:
                    ref_year = "NA"
                Calibration_notes = ref_author + "." + ref_year
            elif "notes" in calibration.keys():
                Calibration_notes = calibration["notes"]
            else:
                Calibration_notes = "NA"
        else:
            Calibration_equation = "NA"
            Calibration_notes = "NA"

        # Truncate the notes if too long
        charlim = 30;
        if len(Calibration_notes) > charlim:
            Calibration_notes = Calibration_notes[0:charlim] + " ..."

        res = {"archiveType": archiveType,
               "authors": authors,
               "Year": Year,
               "DOI": DOI,
               "Variable": Variable,
               "units": units,
               "Climate_Variable": ClimateVar,
               "Detail": Detail,
               "Scope": Scope,
               "Seasonality": Seasonality,
               "Interpretation_Direction": Direction,
               "Calibration_equation": Calibration_equation,
               "Calibration_notes": Calibration_notes}

        return res

    def dashboard(self, figsize=[11, 8], plt_kwargs=None, histplt_kwargs=None, spectral_kwargs=None,
                  spectralsignif_kwargs=None, spectralfig_kwargs=None, map_kwargs=None, metadata=True,
                  savefig_settings=None, ensemble=False, D=None):
        '''

        Parameters
        ----------
        
        figsize : list or tuple, optional

            Figure size. The default is [11,8].

        plt_kwargs : dict, optional

            Optional arguments for the timeseries plot. See Series.plot() or EnsembleSeries.plot_envelope(). The default is None.

        histplt_kwargs : dict, optional

            Optional arguments for the distribution plot. See Series.histplot() or EnsembleSeries.plot_distplot(). The default is None.

        spectral_kwargs : dict, optional

            Optional arguments for the spectral method. Default is to use Lomb-Scargle method. See Series.spectral() or EnsembleSeries.spectral(). The default is None.

        spectralsignif_kwargs : dict, optional

            Optional arguments to estimate the significance of the power spectrum. See PSD.signif_test. Note that we currently do not support significance testing for ensembles. The default is None.

        spectralfig_kwargs : dict, optional

            Optional arguments for the power spectrum figure. See PSD.plot() or MultiplePSD.plot_envelope(). The default is None.

        map_kwargs : dict, optional

            Optional arguments for the map. See LipdSeries.map(). The default is None.

        metadata : bool; {True,False}, optional

            Whether or not to produce a dashboard with printed metadata. The default is True.

        savefig_settings : dict, optional

            the dictionary of arguments for plt.savefig(); some notes below:
            - "path" must be specified; it can be any existed or non-existed path,
              with or without a suffix; if the suffix is not given in "path", it will follow "format"
            - "format" can be one of {"pdf", "eps", "png", "ps"}.
            The default is None.

        ensemble : bool; {True, False}, optional

            If True, will return the dashboard in ensemble modes if ensembles are available

        D : pyleoclim.Lipd object

            If asking for an ensemble plot, a pyleoclim.Lipd object must be provided

        Returns
        -------
        
        fig : matplotlib.figure

            The figure

        ax : matplolib.axis

            The axis

        See also
        --------

        pyleoclim.core.series.Series.plot : plot a timeseries

        pyleoclim.core.ensembleseries.EnsembleSeries.plot_envelope: Envelope plots for an ensemble

        pyleoclim.core.series.Series.histplot : plot a distribution of the timeseries

        pyleoclim.core.ensembleseries.EnsembleSeries.histplot : plot a distribution of the timeseries across ensembles

        pyleoclim.core.series.Series.spectral : spectral analysis method.

        pyleoclim.core.multipleseries.MultipleSeries.spectral : spectral analysis method for multiple series.

        pyleoclim.core.psds.PSD.signif_test : significance test for timeseries analysis

        pyleoclim.core.psds.PSD.plot : plot power spectrum

        pyleoclim.core.psds.MulitplePSD.plot : plot envelope of power spectrum

        pyleoclim.core.lipdseries.LipdSeries.map : map location of dataset

        pyleoclim.core.lipdseries.LipdSeries.getMetadata : get relevant metadata from the timeseries object

        pyleoclim.utils.mapping.map : Underlying mapping function for Pyleoclim

        Examples
        --------

        .. ipython:: python
            :okwarning:
            :okexcept:

            import pyleoclim as pyleo
            url = 'http://wiki.linked.earth/wiki/index.php/Special:WTLiPD?op=export&lipdid=MD982176.Stott.2004'
            data = pyleo.Lipd(usr_path = url)
            ts = data.to_LipdSeries(number=5)
            @savefig ts_dashboard.png
            fig, ax = ts.dashboard()
            pyleo.closefig(fig)

        '''
        if ensemble == True and D is None:
            raise ValueError("When an ensemble dashboard is requested, the corresponsind Lipd object must be supplied")

        if ensemble == True:
            warnings.warn('Some of the computation in ensemble mode can require a few minutes to complete.')

        savefig_settings = {} if savefig_settings is None else savefig_settings.copy()
        res = self.getMetadata()
        # start plotting
        fig = plt.figure(figsize=figsize)
        gs = gridspec.GridSpec(2, 6, wspace=0)
        gs.update(left=0, right=1.1)

        if ensemble == True:
            ens = self.chronEnsembleToPaleo(D)
            ensc = ens.common_time()

        ax = {}
        # Plot the timeseries
        plt_kwargs = {} if plt_kwargs is None else plt_kwargs.copy()
        ax['ts'] = fig.add_subplot(gs[0, :-1])
        plt_kwargs.update({'ax': ax['ts']})
        # use the defaults if color/markers not specified
        if ensemble == False:
            if 'marker' not in plt_kwargs.keys():
                archiveType = lipdutils.LipdToOntology(res['archiveType']).lower().replace(" ", "")
                plt_kwargs.update({'marker': self.plot_default[archiveType][1]})
            if 'color' not in plt_kwargs.keys():
                archiveType = lipdutils.LipdToOntology(res['archiveType']).lower().replace(" ", "")
                plt_kwargs.update({'color': self.plot_default[archiveType][0]})
            ax['ts'] = self.plot(**plt_kwargs)
        elif ensemble == True:
            if 'curve_clr' not in plt_kwargs.keys():
                archiveType = lipdutils.LipdToOntology(res['archiveType']).lower().replace(" ", "")
                plt_kwargs.update({'curve_clr': self.plot_default[archiveType][0]})
            if 'shade_clr' not in plt_kwargs.keys():
                archiveType = lipdutils.LipdToOntology(res['archiveType']).lower().replace(" ", "")
                plt_kwargs.update({'shade_clr': self.plot_default[archiveType][0]})
            # plt_kwargs.update({'ylabel':self.value_name})
            ax['ts'] = ensc.plot_envelope(**plt_kwargs)
        else:
            raise ValueError("Invalid argument value for ensemble")
        ymin, ymax = ax['ts'].get_ylim()

        # plot the distplot
        histplt_kwargs = {} if histplt_kwargs is None else histplt_kwargs.copy()
        ax['dts'] = fig.add_subplot(gs[0, -1:])
        histplt_kwargs.update({'ax': ax['dts']})
        histplt_kwargs.update({'ylabel': 'Counts'})
        histplt_kwargs.update({'vertical': True})
        if 'color' not in histplt_kwargs.keys():
            archiveType = lipdutils.LipdToOntology(res['archiveType']).lower().replace(" ", "")
            histplt_kwargs.update({'color': self.plot_default[archiveType][0]})
        if ensemble == False:
            ax['dts'] = self.histplot(**histplt_kwargs)
        elif ensemble == True:
            ax['dts'] = ensc.histplot(**histplt_kwargs)
        ax['dts'].set_ylim([ymin, ymax])
        ax['dts'].set_yticklabels([])
        ax['dts'].set_ylabel('')
        ax['dts'].set_yticks([])

        # make the map - brute force since projection is not being returned properly
        lat = [self.lipd_ts['geo_meanLat']]
        lon = [self.lipd_ts['geo_meanLon']]

        map_kwargs = {} if map_kwargs is None else map_kwargs.copy()
        if 'projection' in map_kwargs.keys():
            projection = map_kwargs['projection']
        else:
            projection = 'Orthographic'
        if 'proj_default' in map_kwargs.keys():
            proj_default = map_kwargs['proj_default']
        else:
            proj_default = True
        if proj_default == True:
            proj1 = {'central_latitude': lat[0],
                     'central_longitude': lon[0]}
            proj2 = {'central_latitude': lat[0]}
            proj3 = {'central_longitude': lon[0]}
            try:
                proj = mapping.set_proj(projection=projection, proj_default=proj1)
            except:
                try:
                    proj = mapping.set_proj(projection=projection, proj_default=proj3)
                except:
                    proj = mapping.set_proj(projection=projection, proj_default=proj2)
        if 'marker' in map_kwargs.keys():
            marker = map_kwargs['marker']
        else:
            marker = self.plot_default[archiveType][1]
        if 'color' in map_kwargs.keys():
            color = map_kwargs['color']
        else:
            color = self.plot_default[archiveType][0]
        if 'background' in map_kwargs.keys():
            background = map_kwargs['background']
        else:
            background = True
        if 'borders' in map_kwargs.keys():
            borders = map_kwargs['borders']
        else:
            borders = False
        if 'rivers' in map_kwargs.keys():
            rivers = map_kwargs['rivers']
        else:
            rivers = False
        if 'lakes' in map_kwargs.keys():
            lakes = map_kwargs['lakes']
        else:
            lakes = False
        if 'scatter_kwargs' in map_kwargs.keys():
            scatter_kwargs = map_kwargs['scatter_kwargs']
        else:
            scatter_kwargs = {}
        if 'markersize' in map_kwargs.keys():
            scatter_kwargs.update({'s': map_kwargs['markersize']})
        else:
            pass
        if 'lgd_kwargs' in map_kwargs.keys():
            lgd_kwargs = map_kwargs['lgd_kwargs']
        else:
            lgd_kwargs = {}
        if 'legend' in map_kwargs.keys():
            legend = map_kwargs['legend']
        else:
            legend = False
        # make the plot map

        data_crs = ccrs.PlateCarree()
        ax['map'] = fig.add_subplot(gs[1, 0:2], projection=proj)
        ax['map'].coastlines()
        if background is True:
            ax['map'].stock_img()
        # Other extra information
        if borders is True:
            ax['map'].add_feature(cfeature.BORDERS)
        if lakes is True:
            ax['map'].add_feature(cfeature.LAKES)
        if rivers is True:
            ax['map'].add_feature(cfeature.RIVERS)
        ax['map'].scatter(lon, lat, zorder=10, label=marker, facecolor=color, transform=data_crs, **scatter_kwargs)
        if legend == True:
            ax.legend(**lgd_kwargs)

        # spectral analysis
        spectral_kwargs = {} if spectral_kwargs is None else spectral_kwargs.copy()
        if 'method' in spectral_kwargs.keys():
            pass
        else:
            spectral_kwargs.update({'method': 'lomb_scargle'})
        if 'freq_method' in spectral_kwargs.keys():
            pass
        else:
            if ensemble == False:
                spectral_kwargs.update({'freq_method': 'lomb_scargle'})
            elif ensemble == True:
                pass

        ax['spec'] = fig.add_subplot(gs[1, -3:])
        spectralfig_kwargs = {} if spectralfig_kwargs is None else spectralfig_kwargs.copy()
        spectralfig_kwargs.update({'ax': ax['spec']})

        if ensemble == False:
            ts_preprocess = self.detrend().standardize()
            psd = ts_preprocess.spectral(**spectral_kwargs)

            # Significance test
            spectralsignif_kwargs = {} if spectralsignif_kwargs is None else spectralsignif_kwargs.copy()
            psd_signif = psd.signif_test(**spectralsignif_kwargs)
            # plot
            if 'color' not in spectralfig_kwargs.keys():
                archiveType = lipdutils.LipdToOntology(res['archiveType']).lower().replace(" ", "")
                spectralfig_kwargs.update({'color': self.plot_default[archiveType][0]})
            if 'signif_clr' not in spectralfig_kwargs.keys():
                spectralfig_kwargs.update({'signif_clr': 'grey'})
            ax['spec'] = psd_signif.plot(**spectralfig_kwargs)

        elif ensemble == True:
            if 'curve_clr' not in spectralfig_kwargs.keys():
                archiveType = lipdutils.LipdToOntology(res['archiveType']).lower().replace(" ", "")
                spectralfig_kwargs.update({'curve_clr': self.plot_default[archiveType][0]})
            if 'shade_clr' not in spectralfig_kwargs.keys():
                archiveType = lipdutils.LipdToOntology(res['archiveType']).lower().replace(" ", "")
                spectralfig_kwargs.update({'shade_clr': self.plot_default[archiveType][0]})
            psd = ensc.detrend().standardize().spectral(**spectral_kwargs)
            # plot
            ax['spec'] = psd.plot_envelope(**spectralfig_kwargs)

        # Make the plot

        if metadata == True:
            # get metadata
            textstr = "archiveType: " + res["archiveType"] + "\n" + "\n" + \
                      "Authors: " + res["authors"] + "\n" + "\n" + \
                      "Year: " + res["Year"] + "\n" + "\n" + \
                      "DOI: " + res["DOI"] + "\n" + "\n" + \
                      "Variable: " + res["Variable"] + "\n" + "\n" + \
                      "units: " + res["units"] + "\n" + "\n" + \
                      "Climate Interpretation: " + "\n" + \
                      "    Climate Variable: " + res["Climate_Variable"] + "\n" + \
                      "    Detail: " + res["Detail"] + "\n" + \
                      "    Seasonality: " + res["Seasonality"] + "\n" + \
                      "    Direction: " + res["Interpretation_Direction"] + "\n \n" + \
                      "Calibration: \n" + \
                      "    Equation: " + res["Calibration_equation"] + "\n" + \
                      "    Notes: " + res["Calibration_notes"]
            plt.figtext(1.15, 0.4, textstr, fontsize=12)
        
        #gs.tight_layout(fig)

        if 'path' in savefig_settings:
            plotting.savefig(fig, settings=savefig_settings)

        return fig, ax

    def mapNearRecord(self, D, n=5, radius=None, sameArchive=False,
                      projection='Orthographic', proj_default=True,
                      background=True, borders=False, rivers=False,
                      lakes=False, figsize=None, ax=None,
                      marker_ref=None, color_ref=None, marker=None, color=None,
                      markersize_adjust=False, scale_factor=100, scatter_kwargs=None,
                      legend=True, lgd_kwargs=None, savefig_settings=None):
        """ Map records that are near the timeseries of interest

        Parameters
        ----------
        
        D : pyleoclim.Lipd

            A pyleoclim LiPD object

        n : int, optional

            The n number of closest records. The default is 5.

        radius : float, optional

            The radius to take into consideration when looking for records (in km). The default is None.

        sameArchive :  bool; {True, False}, optional

            Whether to consider records from the same archiveType as the original record. The default is False.

        projection : str, optional

            A valid cartopy projection. The default is 'Orthographic'.
            See pyleoclim.utils.mapping for a list of supported projections.

        proj_default : True or dict, optional

            The projection arguments. If not True, then use a dictionary to pass the appropriate arguments depending on the projection. The default is True.

        background : bool; {True, False}, optional

            Whether to use a background. The default is True.

        borders :  bool; {True, False}, optional

            Whether to plot country borders. The default is False.

        rivers :  bool; {True, False}, optional

            Whether to plot rivers. The default is False.

        lakes :  bool; {True, False}, optional

            Whether to plot rivers. The default is False.

        figsize : list or tuple, optional

            the size of the figure. The default is None.

        ax : matplotlib.ax, optional

            The matplotlib axis onto which to return the map. The default is None.

        marker_ref : str, optional

            Marker shape to use for the main record. The default is None, which corresponds to the default marker for the archiveType

        color_ref : str, optional

            The color for the main record. The default is None, which corresponds to the default color for the archiveType.

        marker : str or list, optional

            Marker shape to use for the other records. The default is None, which corresponds to the marker shape for each archiveType.

        color : str or list, optional

            Color for each marker. The default is None, which corresponds to the color for each archiveType

        markersize_adjust :  bool; {True, False}, optional

            Whether to adjust the marker size according to distance from record of interest. The default is False.

        scale_factor : int, optional

            The maximum marker size. The default is 100.

        scatter_kwargs : dict, optional

            Parameters for the scatter plot. The default is None.

        legend :  bool; {True, False}, optional

            Whether to show the legend. The default is True.

        lgd_kwargs : dict, optional

            Parameters for the legend. The default is None.

        savefig_settings : dict, optional

            the dictionary of arguments for plt.savefig(); some notes below:
            - "path" must be specified; it can be any existed or non-existed path,
              with or without a suffix; if the suffix is not given in "path", it will follow "format"
            - "format" can be one of {"pdf", "eps", "png", "ps"}. The default is None.

        Returns
        -------

        res : dict
            contains fig and ax


        See also
        --------

        pyleoclim.utils.mapping.map: Underlying mapping function for Pyleoclim

        pyleoclim.utils.mapping.dist_sphere: Calculate distance on a sphere

        pyleoclim.utils.mapping.compute_dist: Compute the distance between a point and an array

        pyleoclim.utils.mapping.within_distance: Returns point in an array within a certain distance

        """
        scatter_kwargs = {} if scatter_kwargs is None else scatter_kwargs.copy()

        # get the information about the original timeseries
        lat_ref = [self.lipd_ts['geo_meanLat']]
        lon_ref = [self.lipd_ts['geo_meanLon']]

        if 'archiveType' in self.lipd_ts.keys():
            archiveType_ref = lipdutils.LipdToOntology(self.lipd_ts['archiveType']).lower().replace(" ", "")
        else:
            archiveType_ref = 'other'

        # make sure criteria is in the plot_default list
        if archiveType_ref not in self.plot_default.keys():
            archiveType_ref = 'other'

        # get information about the other timeseries
        lat = []
        lon = []
        archiveType = []

        dataSetName_ref = self.lipd_ts['dataSetName']

        for idx, key in enumerate(D.lipd):
            if key != dataSetName_ref:
                d = D.lipd[key]
                lat.append(d['geo']['geometry']['coordinates'][1])
                lon.append(d['geo']['geometry']['coordinates'][0])
                if 'archiveType' in d.keys():
                    archiveType.append(lipdutils.LipdToOntology(d['archiveType']).lower().replace(" ", ""))
                else:
                    archiveType.append('other')

        # make sure criteria is in the plot_default list
        for idx, val in enumerate(archiveType):
            if val not in self.plot_default.keys():
                archiveType[idx] = 'other'

        if len(lat) == 0:  # this should not happen unless the coordinates are not available in the LiPD file
            raise ValueError('no matching record found')

        # Filter by the same type of archive if asked
        if sameArchive == True:
            idx_archive = [idx for idx, val in enumerate(archiveType) if val == archiveType_ref]
            if len(idx_archive) == 0:
                raise ValueError(
                    'No records corresponding to the same archiveType available. Widen your search criteria.')
            else:
                lat = [lat[idx] for idx in idx_archive]
                lon = [lon[idx] for idx in idx_archive]
                archiveType = [archiveType[idx] for idx in idx_archive]

        # compute the distance
        dist = mapping.compute_dist(lat_ref, lon_ref, lat, lon)

        if radius:
            idx_radius = mapping.within_distance(dist, radius)
            if len(idx_radius) == 0:
                raise ValueError('No records withing matching radius distance. Widen your search criteria')
            else:
                lat = [lat[idx] for idx in idx_radius]
                lon = [lon[idx] for idx in idx_radius]
                archiveType = [archiveType[idx] for idx in idx_radius]
                dist = [dist[idx] for idx in idx_radius]

        # print a warning if plotting less than asked because of the filters

        if n > len(dist):
            warnings.warn("Number of matching records is less" + \
                          " than the number of neighbors chosen. Including all records " + \
                          " in the analysis.")
            n = len(dist)

        # Sort the distance array
        sort_idx = np.argsort(dist)
        dist = [dist[idx] for idx in sort_idx]
        lat = [lat[idx] for idx in sort_idx]
        lon = [lon[idx] for idx in sort_idx]
        archiveType = [archiveType[idx] for idx in sort_idx]

        # Grab the right number of records
        dist = dist[0:n]
        lat = lat[0:n]
        lon = lon[0:n]
        archiveType = archiveType[0:n]

        # Get plotting information

        if marker_ref == None:
            marker_ref = self.plot_default[archiveType_ref][1]
        if color_ref == None:
            color_ref = self.plot_default[archiveType_ref][0]

        if marker == None:
            marker = []
            for item in archiveType:
                marker.append(self.plot_default[item][1])
        elif type(marker) == list:
            if len(marker) != len(lon):
                raise ValueError('When providing a list, it should be the same length as the number of records')
        elif type(marker) == str:
            marker = [marker] * len(lon)

        if color == None:
            color = []
            for item in archiveType:
                color.append(self.plot_default[item][0])
        elif type(color) == list:
            if len(color) != len(lon):
                raise ValueError('When providing a list, it should be the same length as the number of records')
        elif type(color) == str:
            color = [color] * len(lon)

        if 'edgecolors' not in scatter_kwargs.keys():
            edgecolors = []
            for item in marker:
                edgecolors.append('w')
            edgecolors.append('k')
            scatter_kwargs.update({'edgecolors': edgecolors})

        # Start plotting
        lat_all = lat + lat_ref
        lon_all = lon + lon_ref
        dist_all = dist + [0]
        archiveType_all = archiveType
        archiveType_all.append(archiveType_ref)

        color_all = color
        color_all.append(color_ref)
        marker_all = marker
        marker_all.append(marker_ref)

        if markersize_adjust == True:
            scale = dist_all[-1] / (scale_factor - 30)
            s = list(np.array(dist_all) * 1 / (scale) + 30)
            s.reverse()
            scatter_kwargs.update({'s': s})

        proj1 = {'central_latitude': lat_ref[0],
                 'central_longitude': lon_ref[0]}
        proj2 = {'central_latitude': lat_ref[0]}
        proj3 = {'central_longitude': lon_ref[0]}

        if proj_default == True:
            try:
                res = mapping.map(lat=lat_all, lon=lon_all,
                                  criteria=archiveType_all,
                                  marker=marker_all, color=color_all,
                                  projection=projection, proj_default=proj1,
                                  background=background, borders=borders,
                                  rivers=rivers, lakes=lakes,
                                  figsize=figsize, ax=ax,
                                  scatter_kwargs=scatter_kwargs, legend=legend,
                                  lgd_kwargs=lgd_kwargs, savefig_settings=savefig_settings)
            except:
                try:
                    res = mapping.map(lat=lat_all, lon=lon_all,
                                      criteria=archiveType_all,
                                      marker=marker_all, color=color_all,
                                      projection=projection, proj_default=proj2,
                                      background=background, borders=borders,
                                      rivers=rivers, lakes=lakes,
                                      figsize=figsize, ax=ax,
                                      scatter_kwargs=scatter_kwargs, legend=legend,
                                      lgd_kwargs=lgd_kwargs, savefig_settings=savefig_settings)
                except:
                    res = mapping.map(lat=lat_all, lon=lon_all,
                                      criteria=archiveType_all,
                                      marker=marker_all, color=color_all,
                                      projection=projection, proj_default=proj3,
                                      background=background, borders=borders,
                                      rivers=rivers, lakes=lakes,
                                      figsize=figsize, ax=ax,
                                      scatter_kwargs=scatter_kwargs, legend=legend,
                                      lgd_kwargs=lgd_kwargs, savefig_settings=savefig_settings)

        else:
            res = mapping.map(lat=lat_all, lon=lon_all,
                              criteria=archiveType_all,
                              marker=marker_all, color=color_all,
                              projection=projection, proj_default=proj_default,
                              background=background, borders=borders,
                              rivers=rivers, lakes=lakes,
                              figsize=figsize, ax=ax,
                              scatter_kwargs=scatter_kwargs, legend=legend,
                              lgd_kwargs=lgd_kwargs, savefig_settings=savefig_settings)

        return res

    def plot_age_depth(self, figsize=[10, 4], plt_kwargs=None,
                       savefig_settings=None,
                       ensemble=False, D=None, num_traces=10, ensemble_kwargs=None,
                       envelope_kwargs=None, traces_kwargs=None):

        '''

        Parameters
        ----------

        figsize : list or tuple, optional

            Size of the figure. The default is [10,4].

        plt_kwargs : dict, optional

            Arguments for basic plot. See Series.plot() for details. The default is None.

        savefig_settings : dict, optional

            the dictionary of arguments for plt.savefig(); some notes below:
            - "path" must be specified; it can be any existed or non-existed path,
              with or without a suffix; if the suffix is not given in "path", it will follow "format"
            - "format" can be one of {"pdf", "eps", "png", "ps"}. The default is None.

        ensemble : bool; {True, False}, optional

            Whether to use age model ensembles stored in the file for the plot. The default is False.
            If no ensemble can be found, will error out.

        D : pyleoclim.Lipd, optional

            The pyleoclim.Lipd object from which the pyleoclim.LipdSeries is derived. The default is None.

        num_traces : int, optional

            Number of individual age models to plot. To plot only the envelope and median value, set this parameter to 0 or None. The default is 10.

        ensemble_kwargs : dict, optional

            Parameters associated with identifying the chronEnsemble tables. See pyleoclim.core.lipdseries.LipdSeries.chronEnsembleToPaleo() for details. The default is None.

        envelope_kwargs : dict, optional

            Parameters to control the envelope plot. See pyleoclim.EnsembleSeries.plot_envelope() for details. The default is None.

        traces_kwargs : dict, optional

            Parameters to control the traces plot. See pyleoclim.EnsembleSeries.plot_traces() for details. The default is None.

        Raises
        ------

        ValueError

            In ensemble mode, make sure that the LiPD object is given

        KeyError

            Depth information needed.

        Returns
        -------

        fig,ax

            The figure

        See also
        --------

        pyleoclim.core.lipd.Lipd : Pyleoclim internal representation of a LiPD file

        pyleoclim.core.series.Series.plot : Basic plotting in pyleoclim

        pyleoclim.core.lipdseries.LipdSeries.chronEnsembleToPaleo : Function to map the ensemble table to a paleo depth.

        pyleoclim.core.ensembleseries.EnsembleSeries.plot_envelope : Create an envelope plot from an ensemble

        pyleoclim.core.ensembleseries.EnsembleSeries.plot_traces : Create a trace plot from an ensemble

        Examples
        --------

        .. ipython:: python
            :okwarning:
            :okexcept:

            D = pyleo.Lipd('http://wiki.linked.earth/wiki/index.php/Special:WTLiPD?op=export&lipdid=Crystal.McCabe-Glynn.2013')
            ts=D.to_LipdSeries(number=2)
            @savefig lipdseries_age_depth.png
            fig, ax = ts.plot_age_depth()
            pyleo.closefig(fig)

        '''
        if ensemble == True and D is None:
            raise ValueError("When an ensemble is requested, the corresponsind Lipd object must be supplied")

        meta = self.getMetadata()
        savefig_settings = {} if savefig_settings is None else savefig_settings.copy()
        plt_kwargs = {} if plt_kwargs is None else plt_kwargs.copy()
        # get depth
        try:
            value_depth, label_depth = lipdutils.checkXaxis(self.lipd_ts, 'depth')
            if 'depthUnits' in self.lipd_ts.keys():
                units_depth = self.lipd_ts['depthUnits']
            else:
                units_depth = 'NA'
        except:
            raise KeyError('No depth available in this record')

        # create a series for which time is actually depth

        if ensemble == False:
            ts = Series(time=self.time, value=value_depth,
                        time_name=self.time_name, time_unit=self.time_unit,
                        value_name=label_depth, value_unit=units_depth)
            plt_kwargs = {} if plt_kwargs is None else plt_kwargs.copy()
            if 'marker' not in plt_kwargs.keys():
                archiveType = lipdutils.LipdToOntology(meta['archiveType']).lower().replace(" ", "")
                plt_kwargs.update({'marker': self.plot_default[archiveType][1]})
            if 'color' not in plt_kwargs.keys():
                archiveType = lipdutils.LipdToOntology(meta['archiveType']).lower().replace(" ", "")
                plt_kwargs.update({'color': self.plot_default[archiveType][0]})

            fig, ax = ts.plot(**plt_kwargs)
        elif ensemble == True:
            ensemble_kwargs = {} if ensemble_kwargs is None else ensemble_kwargs.copy()
            ens = self.chronEnsembleToPaleo(D, **ensemble_kwargs)
            # NOT  A VERY ELEGANT SOLUTION: replace depth in the dictionary
            for item in ens.__dict__['series_list']:
                item.__dict__['value'] = value_depth
                item.__dict__['value_unit'] = units_depth
                item.__dict__['value_name'] = 'depth'
            envelope_kwargs = {} if envelope_kwargs is None else envelope_kwargs.copy()
            if 'curve_clr' not in envelope_kwargs.keys():
                archiveType = lipdutils.LipdToOntology(meta['archiveType']).lower().replace(" ", "")
                envelope_kwargs.update({'curve_clr': self.plot_default[archiveType][0]})
            if 'shade_clr' not in envelope_kwargs.keys():
                archiveType = lipdutils.LipdToOntology(meta['archiveType']).lower().replace(" ", "")
                envelope_kwargs.update({'shade_clr': self.plot_default[archiveType][0]})
            ens2 = ens.common_time()
            if num_traces > 0:
                # envelope_kwargs.update({'mute':True})
                fig, ax = ens2.plot_envelope(**envelope_kwargs)
                traces_kwargs = {} if traces_kwargs is None else traces_kwargs.copy()
                if 'color' not in traces_kwargs.keys():
                    archiveType = lipdutils.LipdToOntology(meta['archiveType']).lower().replace(" ", "")
                    traces_kwargs.update({'color': self.plot_default[archiveType][0]})
                if 'linestyle' not in traces_kwargs.keys():
                    traces_kwargs.update({'linestyle': 'dashed'})
                traces_kwargs.update({'ax': ax, 'num_traces': num_traces})
                ens2.plot_traces(**traces_kwargs)
            else:
                fig, ax = ens2.plot_envelope(**envelope_kwargs)

        if 'fig' in locals():
            if 'path' in savefig_settings:
                plotting.savefig(fig, settings=savefig_settings)
            return fig, ax
        else:
            return ax



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
CorrEns objects store the result of an ensemble correlation calculation between timeseries and/or ensemble of timeseries.
The class enables a print and plot function to easily visualize the result. 
"""

import numpy as np
import seaborn as sns
import pandas as pd
from matplotlib import pyplot as plt, transforms as transforms
from matplotlib.ticker import MaxNLocator
from tabulate import tabulate
from copy import deepcopy

from ..utils import plotting

def pval_format(p, threshold=0.01, style='exp'):
    ''' Print p-value with proper format when p is close to 0
    '''
    if p < threshold:
        if p == 0:
            if style == 'float':
                s = '< 0.000001'
            elif style == 'exp':
                s = '< 1e-6'
            else:
                raise ValueError('Wrong style.')
        else:
            n = int(np.ceil(np.log10(p)))
            if style == 'float':
                s = f'< {10**n}'
            elif style == 'exp':
                s = f'< 1e{n}'
            else:
                raise ValueError('Wrong style.')
    else:
        s = f'{p:.2f}'

    return s

class CorrEns:
    ''' CorrEns objects store the result of an ensemble correlation calculation 
    between timeseries and/or ensemble of timeseries. The class enables a print and 
    plot function to easily visualize the result. 

    Parameters
    ----------

    r: list
    
        the list of correlation coefficients

    p: list
    
        the list of p-values

    p_fmt_td: float
    
        the threshold for p-value formating (0.01 by default, i.e., if p<0.01, will print "< 0.01" instead of "0")

    p_fmt_style: str
    
        the style for p-value formating (exponential notation by default)

    signif: list
    
        the list of significance without FDR

    signif_fdr: list
    
        the list of significance with FDR

    signif_fdr: list
    
        the list of significance with FDR

    alpha : float
    
        The significance level

    See also
    --------

    pyleoclim.utils.correlation.corr_sig : Correlation function
    
    pyleoclim.utils.correlation.fdr : FDR (False Discovery Rate) function
    '''

    def __init__(self, r, p, signif, signif_fdr, alpha, p_fmt_td=0.01, p_fmt_style='exp'):
        self.r = r
        self.p = p
        self.p_fmt_td = p_fmt_td
        self.p_fmt_style = p_fmt_style
        self.signif = signif
        self.signif_fdr = signif_fdr
        self.alpha = alpha
        
    def copy(self):
        '''Copy object
        '''
        return deepcopy(self)

    def __str__(self):
        '''
        Prints out the correlation results
        '''

        pi_list = []
        for pi in self.p:
            pi_list.append(pval_format(pi, threshold=self.p_fmt_td, style=self.p_fmt_style))

        table = {
            'correlation': self.r,
            'p-value': pi_list,
            f'signif. w/o FDR (α: {self.alpha})': self.signif,
            f'signif. w/ FDR (α: {self.alpha})': self.signif_fdr,
        }

        msg = print(tabulate(table, headers='keys'))

        return f'Ensemble size: {len(self.r)}'

    def plot(self, figsize=[4, 4], title=None, ax=None, savefig_settings=None, hist_kwargs=None,
             title_kwargs=None, xlim=None, alpha = 0.8, multiple = 'layer',
             clr_insignif='silver', clr_signif=sns.xkcd_rgb['teal'],
             clr_signif_fdr='darkorange', clr_percentile=sns.xkcd_rgb['salmon']):
        ''' Plot the distribution of correlation values as a histogram
       
        Uses seaborn's `histplot <https://seaborn.pydata.org/generated/seaborn.histplot.html>`_
        
        Color-coding is used to indicate significance, with or without applying 
        the False Discovery Rate (FDR) method. 

        Parameters
        ----------
        
        figsize : list, optional
        
            The figure size. The default is [4, 4].

        title : str, optional
        
            Plot title. The default is None.
            
        multiple: str, optional 
        
            Approach to organizing the 3 different histrograms on the plot. 
            possible values: “layer”[default], “dodge”, “stack”, “fill”
            
        alpha : float in [0, 1]
        
            transparency parameter for histrogram bars. Default: 0.8

        savefig_settings : dict
        
            the dictionary of arguments for plt.savefig(); some notes below:
            - "path" must be specified; it can be any existing or new path,
              with or without a suffix; if the suffix is not given in "path", it will follow "format"
            - "format" can be one of {"pdf", "eps", "png", "ps"}

        hist_kwargs : dict
        
            additional keyword arguments for sns.histplot() [experimental]

        title_kwargs : dict
        
            the keyword arguments for ax.set_title()

        ax : matplotlib.axis, optional
        
            the axis object from matplotlib
            See [matplotlib.axes](https://matplotlib.org/api/axes_api.html) for details.

        xlim : list, optional
        
            x-axis limits. The default is None.

        See also
        --------
        
        pyleoclim.core.series.Series.correlation: correlation with significance
        
        pyleoclim.utils.correlation.fdr: False Discovery Rate

        seaborn.histplot: https://seaborn.pydata.org/generated/seaborn.histplot.html
        
        pyleoclim.utils.plotting.savefig : save figures in Pyleoclim
        '''
        savefig_settings = {} if savefig_settings is None else savefig_settings.copy()
        hist_kwargs = {} if hist_kwargs is None else hist_kwargs.copy()
        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)

        clr_list = [clr_signif_fdr, clr_signif, clr_insignif]
        args = {'multiple': multiple, 'alpha': alpha, 'ax': ax}
        args.update(hist_kwargs)
        
        r_insignif = np.array(self.r)[~np.array(self.signif)]
        r_signif = np.array(self.r)[self.signif]
        r_signif_fdr = np.array(self.r)[self.signif_fdr]
        #r_stack = [r_insignif, r_signif, r_signif_fdr]
        #ax.hist(r_stack, stacked=True, **args)

        # put everything into a dataframe to be able to use seaborn
            
        data = np.empty((len(self.r),3)); data[:] = np.NaN
        col  = [f'p < {self.alpha} (w/ FDR)',f'p < {self.alpha} (w/o FDR)', f'p ≥ {self.alpha}']
        data[self.signif_fdr,0] = r_signif_fdr        
        data[self.signif, 1] = r_signif
        data[~np.array(self.signif),2] = r_insignif

        df = pd.DataFrame(data,columns=col) # place data into a dataframe for seaborn to chew on
        sns.set_palette(sns.color_palette(clr_list)) # update color palette
        ax = sns.histplot(data=df, **args) # draw histogram
        sns.move_legend(ax, "upper left", bbox_to_anchor=(1.08, 1)) # move legend to the right
    
        #ax.legend(loc='upper right', bbox_to_anchor=(1.1, 1), ncol=1)

        frac_signif = np.size(r_signif) / np.size(self.r)
        frac_signif_fdr = np.size(r_signif_fdr) / np.size(self.r)
        ax.text(x=1.1, y=0.4, s=f'Fraction significant: {frac_signif * 100:.1f}%', transform=ax.transAxes, fontsize=10,
                color=clr_signif)
        ax.text(x=1.1, y=0.5, s=f'Fraction significant: {frac_signif_fdr * 100:.1f}%', transform=ax.transAxes,
                fontsize=10, color=clr_signif_fdr)

        r_pcts = np.percentile(self.r, [2.5, 25, 50, 75, 97.5])
        trans = transforms.blended_transform_factory(ax.transData, ax.transAxes)
        for r_pct, pt, ls in zip(r_pcts, np.array([2.5, 25, 50, 75, 97.5]) / 100, [':', '--', '-', '--', ':']):
            ax.axvline(x=r_pct, linestyle=ls, color=clr_percentile)
            ax.text(x=r_pct, y=1.02, s=pt, color=clr_percentile, transform=trans, ha='center', fontsize=10)

        ax.set_xlabel(r'$r$')
        ax.set_ylabel('Count')
        ax.yaxis.set_major_locator(MaxNLocator(integer=True))

        if xlim is not None:
            ax.set_xlim(xlim)

        if title is not None:
            title_kwargs = {} if title_kwargs is None else title_kwargs.copy()
            t_args = {'y': 1.1, 'weight': 'bold'}
            t_args.update(title_kwargs)
            ax.set_title(title, **t_args)

        if 'fig' in locals():
            if 'path' in savefig_settings:
                plotting.savefig(fig, settings=savefig_settings)
            return fig, ax
        else:
            return ax




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This module converts Pyleoclim objects to and from JSON files. 

Useful for obtaining a human-readable output and keeping the results of an analysis. The JSON file can also be used to swap analysis results between programming languages. 

These utilities are maintained on an as-needed basis and that not all objects are currently available.
"""

__all__ =['PyleoObj_to_json', 'json_to_PyleoObj', 'isPyleoclim']

import pyleoclim as pyleo
import numpy as np
import inspect
import json
from urllib.request import urlopen
import re

def isPyleoclim(obj):
    '''
    Check whether an object is a valid type for Pyleoclim ui object

    Parameters
    ----------
    obj : pyleoclim.core.ui
        Object from the Pyleoclim UI module

    Returns
    -------
    Bool : {True,False}
        

    '''
    class_names=[]
    for name, ob in inspect.getmembers(pyleo.core):
            if inspect.isclass(ob):
                class_names.append(ob)
    return type(obj) in class_names

def PyleoObj_to_dict(obj):
    '''Transform a pyleoclim object into a dictionary. 
    
    The transformation ensures that all the objects are JSON serializable (i.e. all numpy arrays have been converted to lists.)
    

    Parameters
    ----------
    obj : pyleoclim.core.io
        A pyleoclim object from the UI module

    Returns
    -------
    s : dict
        A JSON-encodable dictionary
        
    See also
    --------
    
    pyleoclim.utils.jsonutils.isPyleoclim : Whether an object is a valid Pyleoclim object

    '''
    
    if isinstance(obj,(dict)):
        s=obj
    else:
        s=vars(obj)
    for k in s.keys():
        #print(k)
        if isinstance(s[k],(np.ndarray)):            
            s[k] = s[k].astype('float64').tolist()
        elif isinstance(s[k],(dict)):
            s[k]=PyleoObj_to_dict(s[k])
        elif isPyleoclim(s[k])==True:
            s[k]=PyleoObj_to_dict(s[k])
        elif isinstance(s[k],(list)):
            if isPyleoclim(s[k][0])==True:
                new_list=[]
                for item in s[k]:
                    new_list.append(PyleoObj_to_dict(item))
                s[k]=new_list                    
    
    return s

def PyleoObj_to_json(obj, filename):
    '''Serializes a Pyleoclim object into a JSON file

    Parameters
    ----------
    obj : pyleoclim.core.ui
        A Pyleoclim object from the UI module
        
    filename : str
        Filename or path to save the JSON to.

    Returns
    -------
    None 
    
    See also
    --------
    pyleoclim.utils.jsonutils.PyleoObj_to_dict : Encodes a Pyleoclim UI object into a dictionary that is JSON serializable
    
    '''
    
    s = PyleoObj_to_dict(obj)
    with open(filename,'w') as f:
        json.dump(s, f)
        f.close()
    

def open_json(filename):
    '''Open a json file.

    Parameters
    ----------
    filename : str
        Path to the json file or URL

    Returns
    -------
    t : dict
        A Python dictionary from the json

    '''
    
    regex = re.compile(
        r'^(?:http|ftp)s?://' # http:// or https://
        r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|' #domain...
        r'localhost|' #localhost...
        r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})' # ...or ip
        r'(?::\d+)?' # optional port
        r'(?:/?|[/?]\S+)$', re.IGNORECASE)
    
    if (re.match(regex, filename) is not None)==True:
        response = urlopen(filename)
        t = json.loads(response.read())
    else:    
        with open(filename,'r') as f:
            t=json.load(f)
    return t

def objname_to_obj(objname):
    '''Returns the correct obj type for the name of a Pyleoclim UI object

    Parameters
    ----------
    objname : str
        Name of the object (e.g., Series, Scalogram, MultipleSeries...)

    Raises
    ------
    ValueError
        If the name of the object is not valid

    Returns
    -------
    obj : pyleoclim.core.ui
        A valid Pyleoclim object for the UI module

    '''
    
    possible_objects={'Series':pyleo.core.Series,
                      'PSD':pyleo.core.PSD,
                      'Scalogram':pyleo.core.Scalogram,
                      'Coherence':pyleo.core.Coherence,
                      'MultipleSeries':pyleo.core.MultipleSeries,
                      'SurrogateSeries':pyleo.core.SurrogateSeries,
                      'EnsembleSeries':pyleo.core.EnsembleSeries,
                      'MultiplePSD':pyleo.core.MultiplePSD,
                      'MultipleScalogram':pyleo.core.MultipleScalogram,
                      'Corr':pyleo.core.Corr,
                      'CorrEns':pyleo.core.CorrEns,
                      'SpatialDecomp':pyleo.core.SpatialDecomp,
                      'SsaRes':pyleo.core.SsaRes,
                      'Lipd':pyleo.core.Lipd,
                      'LipdSeries':pyleo.core.LipdSeries
        }
    
    try:
        obj=possible_objects[objname]
    except:
        raise ValueError("The object is not a proper Pyleoclim object")
        
    return obj

def json_to_PyleoObj(filename,objname):
    '''Reads a JSON serialized file into a Pyleoclim object

    Parameters
    ----------
    filename : str
        The filename/path/URL of the JSON-serialized object
    objname : str
        Name of the object (e.g., Series, Scalogram, MultipleSeries...)

    Returns
    -------
    pyleoObj : pyleoclim.core.ui
        A Pyleoclim UI object
        
    See also
    --------
    pyleoclim.utils.jsonutils.open_json : open a json file from a local source or URL
    
    pyleoclim.utils.jsonutils.objename_to_obj : create a valid Pyleoclim object from a string   

    '''
    
    obj = objname_to_obj(objname)
    a = open_json(filename)

    for k in a.keys():
        if k == 'timeseries' or k == 'timeseries1' or k == 'timeseries2':
            try:
                a[k]=pyleo.Series(**a[k])
            except:
                # get rid of LiPD
                del a[k]['plot_default']
                del a[k]['lipd_ts']
                a[k]=pyleo.Series(**a[k])
        if k == 'psd_list':
            for idx,item in enumerate(a[k]):
                if item['timeseries'] is not None:
                    item['timeseries'] = pyleo.Series(**item['timeseries'])
                a[k][idx] = pyleo.PSD(**a[k][idx])
        if k == 'scalogram_list':
            for idx,item in enumerate(a[k]):
                if item['timeseries'] is not None:
                    item['timeseries'] = pyleo.Series(**item['timeseries'])
                a[k][idx] = pyleo.Scalogram(**a[k][idx])
        if k == 'signif_qs' and a[k] is not None:
            if obj==pyleo.core.PSD:
                for idx,item in enumerate(a[k]['psd_list']):
                    if item['timeseries'] is not None:
                        item['timeseries'] = pyleo.Series(**item['timeseries'])
                    a[k]['psd_list'][idx]=pyleo.PSD(**a[k]['psd_list'][idx])
                a[k] = pyleo.MultiplePSD(**a[k])
            elif obj == pyleo.core.Scalogram:
                for idx,item in enumerate(a[k]['scalogram_list']):
                    if item['timeseries'] is not None:
                        item['timeseries'] = pyleo.Series(**item['timeseries'])
                    a[k]['scalogram_list'][idx]=pyleo.Scalogram(**a[k]['scalogram_list'][idx])
                a[k] = pyleo.MultipleScalogram(**a[k])
            elif obj == pyleo.core.Coherence:
                 for idx,item in enumerate (a[k]):
                     for idx2,item2 in enumerate(item['scalogram_list']):
                         if item2['timeseries'] is not None:
                             item2['timeseries'] = pyleo.Series(**item2['timeseries'])
                         item['scalogram_list'][idx2]=pyleo.Scalogram(**item['scalogram_list'][idx2])
                     a[k][idx] = pyleo.MultipleScalogram(**a[k][idx])           
        if k == 'signif_scals' and a[k] is not None:
            for idx,item in enumerate(a[k]['scalogram_list']):
                if item['timeseries'] is not None:
                    item['timeseries'] = pyleo.Series(**item['timeseries'])
                a[k]['scalogram_list'][idx]=pyleo.Scalogram(**a[k]['scalogram_list'][idx])
            a[k] = pyleo.MultipleScalogram(**a[k])     

    pyleoObj=obj(**a)
    
    return pyleoObj